# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/10/14 22:54
describe: 命令行工具集

https://click.palletsprojects.com/en/8.0.x/quickstart/
"""
import click


@click.group()
def czsc():
    """CZSC 命令行工具"""
    pass


@czsc.command()
def a():
    """随机输出一条缠中说禅良言警句"""
    from czsc.aphorism import print_one
    print_one()



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/6/24 18:49
describe: Tushare数据源
"""
import os
import czsc
import pandas as pd
from czsc import Freq, RawBar
from typing import List
from tqdm import tqdm
from loguru import logger

# 首次使用需要打开一个python终端按如下方式设置 token
# czsc.set_url_token(token='your token', url='http://api.tushare.pro')

cache_path = os.getenv("TS_CACHE_PATH", os.path.expanduser("~/.ts_data_cache"))
dc = czsc.DataClient(url="http://api.tushare.pro", cache_path=cache_path)


def format_kline(kline: pd.DataFrame, freq: Freq) -> List[RawBar]:
    """Tushare K线数据转换

    :param kline: Tushare 数据接口返回的K线数据
    :param freq: K线周期
    :return: 转换好的K线数据
    """
    bars = []
    dt_key = "trade_time" if "分钟" in freq.value else "trade_date"
    kline = kline.sort_values(dt_key, ascending=True, ignore_index=True)
    records = kline.to_dict("records")

    for i, record in enumerate(records):
        if freq == Freq.D:
            vol = int(record["vol"] * 100) if record["vol"] > 0 else 0
            amount = int(record.get("amount", 0) * 1000)
        else:
            vol = int(record["vol"]) if record["vol"] > 0 else 0
            amount = int(record.get("amount", 0))

        # 将每一根K线转换成 RawBar 对象
        bar = RawBar(
            symbol=record["ts_code"],
            dt=pd.to_datetime(record[dt_key]),
            id=i,
            freq=freq,
            open=record["open"],
            close=record["close"],
            high=record["high"],
            low=record["low"],
            vol=vol,  # 成交量，单位：股
            amount=amount,  # 成交额，单位：元
        )
        bars.append(bar)
    return bars


def get_sw_members(level="L1"):
    """获取申万行业分类成分股"""
    # https://tushare.pro/document/2?doc_id=181 申万行业分类
    sw = dc.index_classify(level=level, src="SW2021")
    rows = []
    for _index in sw.to_dict(orient="records"):
        # https://tushare.pro/document/2?doc_id=182 申万行业成分股
        dfp = dc.index_member(index_code=_index["index_code"])
        dfp["industry_name"] = _index["industry_name"]
        dfp = dfp[dfp["is_new"] == "Y"]
        rows.append(dfp)
    df = pd.concat(rows, ignore_index=True)
    return df


def get_daily_basic(sdt="20100101", edt="20240101"):
    """获取全市场A股的每日指标数据

    https://s0cqcxuy3p.feishu.cn/wiki/W5UYwk0qwiHWlxk2ngDcjxQKncg
    """
    sdt = pd.to_datetime(sdt).strftime("%Y%m%d")
    edt = pd.to_datetime(edt).strftime("%Y%m%d")

    dates = czsc.get_trading_dates(sdt, edt)
    rows = []
    for date in tqdm(dates, desc="获取全市场A股的每日指标"):
        try:
            date = pd.to_datetime(date).strftime("%Y%m%d")
            df = dc.daily_basic(trade_date=date)
            rows.append(df)
        except Exception as e:
            logger.error(f"{date} {e}")
    dfb = pd.concat(rows, ignore_index=True)
    return dfb


def moneyflow_hsgt(start_date, end_date):
    """获取沪深港通资金流向数据

    https://tushare.pro/document/2?doc_id=47

    :param start_date: str, 开始日期, 格式为 "YYYYMMDD"
    :param end_date: str, 结束日期, 格式为 "YYYYMMDD"
    :return: DataFrame, 包含沪深港通资金流向数据
    """
    sdt = pd.to_datetime(start_date)
    edt = pd.to_datetime(end_date)
    dts = pd.date_range(sdt, edt, freq="YE").to_list() + [sdt, edt]
    dts = sorted(list(set(dts)))

    rows = []
    for dt1, dt2 in zip(dts[:-1], dts[1:]):
        ttl = -1 if dt2 != edt else 0
        df1 = dc.moneyflow_hsgt(
            start_date=dt1.strftime("%Y%m%d"),
            end_date=dt2.strftime("%Y%m%d"),
            fields="trade_date, ggt_ss, ggt_sz, hgt, sgt, north_money, south_money",
            ttl=ttl,
        )
        df1 = df1.fillna(0)
        if not df1.empty:
            rows.append(df1)

    df = pd.concat(rows, axis=0, ignore_index=True)
    df["dt"] = pd.to_datetime(df["trade_date"])
    df.sort_values("dt", inplace=True)
    df.drop_duplicates(subset=["dt"], keep="last", inplace=True)
    df = df.drop(["trade_date"], axis=1).reset_index(drop=True)
    df.fillna(0, inplace=True)
    return df


def get_symbols(step="all"):
    """获取标的代码"""
    stocks = dc.stock_basic(exchange="", list_status="L", fields="ts_code,symbol,name,area,industry,list_date")
    stocks_ = stocks[stocks["list_date"] < "2010-01-01"].ts_code.to_list()
    stocks_map = {
        "index": [
            "000905.SH",
            "000016.SH",
            "000300.SH",
            "000001.SH",
            "000852.SH",
            "399001.SZ",
            "399006.SZ",
            "399376.SZ",
            "399377.SZ",
            "399317.SZ",
            "399303.SZ",
        ],
        "stock": stocks.ts_code.to_list(),
        "check": ["000001.SZ"],
        "train": stocks_[:200],
        "valid": stocks_[200:600],
        "etfs": [
            "512880.SH",
            "518880.SH",
            "515880.SH",
            "513050.SH",
            "512690.SH",
            "512660.SH",
            "512400.SH",
            "512010.SH",
            "512000.SH",
            "510900.SH",
            "510300.SH",
            "510500.SH",
            "510050.SH",
            "159992.SZ",
            "159985.SZ",
            "159981.SZ",
            "159949.SZ",
            "159915.SZ",
        ],
    }

    asset_map = {"index": "I", "stock": "E", "check": "E", "train": "E", "valid": "E", "etfs": "FD"}

    if step.lower() == "all":
        symbols = []
        for k, v in stocks_map.items():
            symbols += [f"{ts_code}#{asset_map[k]}" for ts_code in v]
    else:
        asset = asset_map[step]
        symbols = [f"{ts_code}#{asset}" for ts_code in stocks_map[step]]

    return symbols


def get_raw_bars(symbol, freq, sdt, edt, fq="后复权", raw_bar=True):
    """读取本地数据"""
    from czsc import data

    tdc = data.TsDataCache(data_path=cache_path)
    ts_code, asset = symbol.split("#")
    freq = str(freq)
    adj = "qfq" if fq == "前复权" else "hfq"

    if "分钟" in freq:
        freq = freq.replace("分钟", "min")
        bars = tdc.pro_bar_minutes(ts_code, sdt=sdt, edt=edt, freq=freq, asset=asset, adj=adj, raw_bar=raw_bar)

    else:
        _map = {"日线": "D", "周线": "W", "月线": "M"}
        freq = _map[freq]
        bars = tdc.pro_bar(ts_code, start_date=sdt, end_date=edt, freq=freq, asset=asset, adj=adj, raw_bar=raw_bar)
    return bars



"""
技术指标因子
"""
import inspect
import hashlib
import pandas as pd


def CCF(df, **kwargs):
    """使用 CZSC 库中的 factor 识别因子，主要用于识别缠论/形态因子

    :param df: 标准K线数据，DataFrame结构
    :param kwargs: 其他参数

        - czsc_factor: dict, 缠论因子配置，样例：

            {
                "signals_all": ["日线_D1_表里关系V230101_向上_任意_任意_0"],
                "signals_any": [],
                "signals_not": ["日线_D1_涨跌停V230331_涨停_任意_任意_0"],
            }

        - freq: str, default '日线'，K线级别
        - tag: str, default None，标签，用于区分不同的因子

    :return: pd.DataFrame
    """
    from czsc.objects import Factor
    from czsc.utils import format_standard_kline
    from czsc.traders.base import generate_czsc_signals
    from czsc.traders.sig_parse import get_signals_config

    czsc_factor = kwargs.get('czsc_factor', None)
    freq = kwargs.get('freq', '日线')
    assert czsc_factor is not None and isinstance(czsc_factor, dict), "factor 参数必须指定"
    tag = kwargs.get('tag', hashlib.sha256(f"{czsc_factor}_{freq}".encode()).hexdigest().upper()[:6])

    factor_name = inspect.stack()[0][3]
    factor_col = f'F#{factor_name}#{tag}'

    czsc_factor = Factor.load(czsc_factor)
    signals_seq = czsc_factor.signals_all + czsc_factor.signals_any + czsc_factor.signals_not
    signals_config = get_signals_config([x.signal for x in signals_seq])

    bars = format_standard_kline(df, freq=freq)
    dfs = generate_czsc_signals(bars, signals_config, init_n=300, sdt=bars[0].dt, df=True)
    dfs[factor_col] = dfs.apply(czsc_factor.is_match, axis=1).astype(int)

    df = pd.merge(df, dfs[['dt', factor_col]], on='dt', how='left')
    df[factor_col] = df[factor_col].fillna(0)
    return df



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/2/16 20:31
describe: czsc.utils 单元测试
"""
import sys
import pytest
import time
import pandas as pd
import numpy as np
from czsc import utils
from czsc.utils import timeout_decorator


def test_x_round():
    assert utils.x_round(100, 3) == 100
    assert utils.x_round(1.000342, 3) == 1.0
    assert utils.x_round(1.000342, 4) == 1.0003
    assert utils.x_round(1.000342, 5) == 1.00034


def test_fernet():
    from czsc.utils.fernet import generate_fernet_key, fernet_encrypt, fernet_decrypt

    key = generate_fernet_key()
    text = {"account": "admin", "password": "123456"}
    encrypted = fernet_encrypt(text, key)
    decrypted = fernet_decrypt(encrypted, key, is_dict=True)
    assert text == decrypted, f"{text} != {decrypted}"


def test_subtract_fee():
    from czsc.utils.stats import subtract_fee

    # 构造测试数据
    data = {
        "dt": pd.date_range("2022-01-01", periods=20, freq="D"),
        "pos": [0, 1, 1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 1, 0, 0, -1, -1, 0, 1, 1],
        "price": [10, 11, 12, 13, 14, 10, 11, 12, 13, 14, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
    }
    df = pd.DataFrame(data)

    # 执行函数
    df = subtract_fee(df, fee=100)

    # 验证结果
    assert int(df["edge_pre_fee"].sum()) == 2748
    assert int(df["edge_post_fee"].sum()) == 1848

    # 构造测试数据
    data = {
        "dt": pd.date_range("2022-01-01", periods=5, freq="D"),
        "pos": [0, 1, 1, -1, 0],
    }
    df = pd.DataFrame(data)

    # 执行函数并捕获异常
    with pytest.raises(AssertionError):
        subtract_fee(df, fee=1)

    # 构造测试数据
    data = {
        "dt": pd.date_range("2022-01-01", periods=5, freq="D"),
        "pos": [0, 1, 2, -1, 0],
        "price": [10, 11, 12, 13, 14],
    }
    df = pd.DataFrame(data)

    # 执行函数并捕获异常
    with pytest.raises(AssertionError):
        subtract_fee(df, fee=1)


def test_ranker():
    import numpy as np
    import pandas as pd
    from czsc.utils.cross import cross_sectional_ranker

    np.random.seed(42)
    dates = pd.date_range("2021-01-01", "2023-01-05")
    symbols = ["AAPL", "GOOG", "TSLA", "MSFT"]
    data = {"date": [], "symbol": [], "return": [], "factor1": [], "factor2": []}
    for date in dates:
        returns = np.random.randn(len(symbols))
        ranks = np.argsort(returns) + 1
        for ticker, rank in zip(symbols, ranks):
            data["date"].append(date)
            data["symbol"].append(ticker)
            data["return"].append(rank)  # 'return' 现在代表了每天的收益率排名
            data["factor1"].append(np.random.randn())
            data["factor2"].append(np.random.randn())
    df = pd.DataFrame(data)
    df["dt"] = df["date"]

    x_cols = ["factor1", "factor2"]
    y_col = "return"

    dfp = cross_sectional_ranker(df, x_cols, y_col)
    assert dfp["rank"].max() == len(symbols)
    assert dfp["rank"].min() == 1
    assert dfp["rank"].mean() == 2.5


def test_daily_performance():
    from czsc.utils.stats import daily_performance

    # Test case 1: empty daily returns
    result = daily_performance([])
    assert result == {
        "绝对收益": 0,
        "年化": 0,
        "夏普": 0,
        "最大回撤": 0,
        "卡玛": 0,
        "日胜率": 0,
        "日盈亏比": 0,
        "日赢面": 0,
        "年化波动率": 0,
        "下行波动率": 0,
        "非零覆盖": 0,
        "盈亏平衡点": 0,
        "新高间隔": 0,
        "新高占比": 0,
        "回撤风险": 0,
    }

    # Test case 2: daily returns with zero standard deviation
    result = daily_performance([1, 1, 1, 1, 1])
    assert result == {
        "绝对收益": 0,
        "年化": 0,
        "夏普": 0,
        "最大回撤": 0,
        "卡玛": 0,
        "日胜率": 0,
        "日盈亏比": 0,
        "日赢面": 0,
        "年化波动率": 0,
        "下行波动率": 0,
        "非零覆盖": 0,
        "盈亏平衡点": 0,
        "新高间隔": 0,
        "新高占比": 0,
        "回撤风险": 0,
    }

    # Test case 3: daily returns with all zeros
    result = daily_performance([0, 0, 0, 0, 0])
    assert result == {
        "绝对收益": 0,
        "年化": 0,
        "夏普": 0,
        "最大回撤": 0,
        "卡玛": 0,
        "日胜率": 0,
        "日盈亏比": 0,
        "日赢面": 0,
        "年化波动率": 0,
        "下行波动率": 0,
        "非零覆盖": 0,
        "盈亏平衡点": 0,
        "新高间隔": 0,
        "新高占比": 0,
        "回撤风险": 0,
    }

    # Test case 4: normal daily returns
    daily_returns = np.array([0.01, 0.02, -0.01, 0.03, 0.02, -0.02, 0.01, -0.01, 0.02, 0.01])
    result = daily_performance(daily_returns)
    assert result == {
        "绝对收益": 0.08,
        "年化": 2.016,
        "夏普": 5,
        "最大回撤": 0.02,
        "卡玛": 10,
        "日胜率": 0.7,
        "日盈亏比": 1.2857,
        "日赢面": 0.6,
        "年化波动率": 0.2439,
        "下行波动率": 0.0748,
        "非零覆盖": 1.0,
        "盈亏平衡点": 0.7,
        "新高间隔": 5,
        "新高占比": 0.6,
        "回撤风险": 0.082,
    }

    # Test case 5: normal daily returns with different input type
    result = daily_performance([0.01, 0.02, -0.01, 0.03, 0.02, -0.02, 0.01, -0.01, 0.02, 0.01])
    assert result == {
        "绝对收益": 0.08,
        "年化": 2.016,
        "夏普": 5,
        "最大回撤": 0.02,
        "卡玛": 10,
        "日胜率": 0.7,
        "日盈亏比": 1.2857,
        "日赢面": 0.6,
        "年化波动率": 0.2439,
        "下行波动率": 0.0748,
        "非零覆盖": 1.0,
        "盈亏平衡点": 0.7,
        "新高间隔": 5,
        "新高占比": 0.6,
        "回撤风险": 0.082,
    }


def test_find_most_similarity():
    from czsc.utils.features import find_most_similarity

    # 创建一个向量和一个矩阵
    vector = pd.Series(np.random.rand(10))
    matrix = pd.DataFrame(np.random.rand(10, 100))

    # 调用函数
    result = find_most_similarity(vector, matrix, n=5, metric="cosine")

    # 检查结果的类型
    assert isinstance(result, pd.Series)

    # 检查结果的长度im
    assert len(result) == 5

    # 检查结果的索引
    assert all(isinstance(index, int) for index in result.index)

    # 检查结果的值
    assert all(0 <= value <= 1 for value in result.values)


def test_overlap():
    from czsc.utils import overlap

    # 创建一个测试 DataFrame
    df = pd.DataFrame(
        {
            "dt": pd.date_range(start="1/1/2022", periods=5),
            "symbol": ["AAPL", "AAPL", "AAPL", "AAPL", "AAPL"],
            "col": [1, 1, 2, 2, 1],
        }
    )

    # 调用 overlap 函数
    result = overlap(df, "col")

    # 验证结果
    assert result["col_overlap"].tolist() == [1, 2, 1, 2, 1]


def test_timeout_decorator_success():
    @timeout_decorator(2)
    def fast_function():
        time.sleep(1)
        return "Completed"

    assert fast_function() == "Completed"


def test_timeout_decorator_timeout():
    @timeout_decorator(1)
    def slow_function():
        time.sleep(5)
        return "Completed"

    assert slow_function() is None



import pytest
import pandas as pd
from czsc.eda import weights_simple_ensemble


def test_cal_yearly_days():
    if pd.__version__ < "2.1.0":
        pytest.skip("skip this test if pandas version is less than 1.3.0")

    from czsc.eda import cal_yearly_days

    # Test with a list of dates within a single year
    dts = ["2023-01-01", "2023-01-02", "2023-01-03", "2023-12-31"]
    assert cal_yearly_days(dts) == 252

    # Test with a list of dates spanning more than one year
    dts = ["2022-01-01", "2022-12-31", "2023-01-01", "2023-12-31"]
    assert cal_yearly_days(dts) == 2

    # Test with a list of dates with minute precision
    dts = [
        "2023-01-01 12:00",
        "2023-01-02 13:00",
        "2023-01-01 14:00",
        "2023-02-01 15:00",
        "2023-03-01 16:00",
        "2023-03-01 17:00",
    ]
    assert cal_yearly_days(dts) == 252

    # Test with an empty list
    with pytest.raises(AssertionError):
        cal_yearly_days([])

    # Test with a list of dates with duplicates
    dts = ["2023-01-01", "2023-01-01", "2023-01-02", "2023-01-02"]
    assert cal_yearly_days(dts) == 252


def test_weights_simple_ensemble_mean():
    df = pd.DataFrame({"strategy1": [0.1, 0.2, 0.3], "strategy2": [0.2, 0.3, 0.4], "strategy3": [0.3, 0.4, 0.5]})
    weight_cols = ["strategy1", "strategy2", "strategy3"]
    result = weights_simple_ensemble(df, weight_cols, method="mean")
    expected = pd.Series([0.2, 0.3, 0.4], name="weight")
    pd.testing.assert_series_equal(result["weight"], expected)


def test_weights_simple_ensemble_vote():
    df = pd.DataFrame({"strategy1": [1, -1, 1], "strategy2": [-1, 1, -1], "strategy3": [1, 1, -1]})
    weight_cols = ["strategy1", "strategy2", "strategy3"]
    result = weights_simple_ensemble(df, weight_cols, method="vote")
    expected = pd.Series([1, 1, -1], name="weight")
    pd.testing.assert_series_equal(result["weight"], expected)


def test_weights_simple_ensemble_sum_clip():
    df = pd.DataFrame({"strategy1": [0.5, -0.5, 0.5], "strategy2": [0.5, 0.5, -0.5], "strategy3": [0.5, 0.5, 0.5]})
    weight_cols = ["strategy1", "strategy2", "strategy3"]
    result = weights_simple_ensemble(df, weight_cols, method="sum_clip", clip_min=-1, clip_max=1)
    expected = pd.Series([1, 0.5, 0.5], name="weight")
    pd.testing.assert_series_equal(result["weight"], expected)


def test_weights_simple_ensemble_only_long():
    df = pd.DataFrame({"strategy1": [0.5, -0.5, 0.5], "strategy2": [0.5, 0.5, -0.5], "strategy3": [0.5, 0.5, 0.5]})
    weight_cols = ["strategy1", "strategy2", "strategy3"]
    result = weights_simple_ensemble(df, weight_cols, method="sum_clip", clip_min=-1, clip_max=1, only_long=True)
    expected = pd.Series([1, 0.5, 0.5], name="weight")
    pd.testing.assert_series_equal(result["weight"], expected)


def test_limit_leverage():
    from czsc.eda import limit_leverage

    data = {
        "dt": pd.date_range(start="2023-01-01", periods=10, freq="D"),
        "symbol": ["TEST"] * 10,
        "weight": [0.1, 0.2, -0.3, 3, -0.5, 0.6, -0.7, 0.8, -0.9, 1.0],
        "price": [100 + i for i in range(10)],
    }
    df = pd.DataFrame(data)

    # Test with leverage = 1.0
    df_result = limit_leverage(df, leverage=1.0, copy=True, window=3, min_periods=2)
    assert df_result["weight"].max() <= 1.0
    assert df_result["weight"].min() >= -1.0

    # Test with leverage = 2.0
    df_result = limit_leverage(df, leverage=2.0, copy=True, window=3, min_periods=2)
    assert df_result["weight"].max() <= 2.0
    assert df_result["weight"].min() >= -2.0

    # Test with different window and min_periods
    df_result = limit_leverage(df, leverage=1.0, window=5, min_periods=2, copy=True)
    assert df_result["weight"].max() <= 1.0
    assert df_result["weight"].min() >= -1.0

    df1 = df.copy()
    df1.rename(columns={"weight": "weight1"}, inplace=True)
    # Test with leverage = 1.0
    df_result = limit_leverage(df1, leverage=1.0, copy=True, window=3, min_periods=2, weight="weight1")
    assert df_result["weight1"].max() <= 1.0
    assert df_result["weight1"].min() >= -1.0


if __name__ == "__main__":
    pytest.main()



import os
import pandas as pd
from czsc.utils.cache import disk_cache, home_path, empty_cache_path

empty_cache_path()
temp_path = os.path.join(home_path, "temp")


# Create a simple function for testing
@disk_cache(path=temp_path, suffix="pkl", ttl=100)
def run_func_x(x):
    return x * 2


@disk_cache(path=temp_path, suffix="txt", ttl=100)
def run_func_text(x):
    return f"hello {x}"


@disk_cache(path=temp_path, suffix="json", ttl=100)
def run_func_json(x):
    return {"a": 1, "b": 2, "x": x}


@disk_cache(path=temp_path, suffix="xlsx", ttl=100)
def run_func_y(x):
    df = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6], 'x': [x, x, x]})
    return df


@disk_cache(path=temp_path, suffix="feather", ttl=100)
def run_feather(x):
    df = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6], 'x': [x, x, x]})
    return df


@disk_cache(path=temp_path, suffix="parquet", ttl=100)
def run_parquet(x):
    df = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6], 'x': [x, x, x]})
    return df


def test_disk_cache():
    # Call the function
    result = run_func_x(5)

    # Check if the output is correct
    assert result == 10

    # Call the function again with the same argument
    result = run_func_x(5)

    # Check if the output is still correct
    assert result == 10

    # Call the function with a different argument
    result = run_func_text(6)
    result = run_func_text(6)
    assert result == "hello 6"

    # Call the function with a different argument
    result = run_func_json(7)
    result = run_func_json(7)
    assert result == {"a": 1, "b": 2, "x": 7}

    result = run_feather(8)
    result = run_feather(8)
    assert isinstance(result, pd.DataFrame)

    result = run_parquet(9)
    result = run_parquet(9)
    assert isinstance(result, pd.DataFrame)

    # Check if the cache file exists
    files = os.listdir(os.path.join(temp_path, "run_func_x"))
    assert len(files) == 1

    # Call the function with a different argument
    result = run_func_y(5)
    files = os.listdir(os.path.join(temp_path, "run_func_y"))
    assert len(files) == 1
    file_xlsx = [x for x in files if x.endswith("xlsx")][0]
    df = pd.read_excel(os.path.join(temp_path, f"run_func_y/{file_xlsx}"))
    assert isinstance(df, pd.DataFrame)



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2024/07/11 12:39
describe: Fernet 加密解密
"""
import os
from typing import Union


def generate_fernet_key():
    """生成 Fernet key

    等价于：base64.urlsafe_b64encode(os.urandom(32))
    """
    from cryptography.fernet import Fernet

    key = Fernet.generate_key()
    return key.decode()


def fernet_encrypt(data: Union[dict, str], key: str = None) -> str:
    """加密文本/字典

    :param data: 需要加密的文本、字典
    :param key: Fernet key must be 32 url-safe base64-encoded bytes.
        推荐使用 generate_fernet_key() 生成
    :return: 加密后的文本
    """
    from cryptography.fernet import Fernet

    key = key or os.getenv("FERNET_KEY")
    cipher_suite = Fernet(key.encode())
    encrypted_text = cipher_suite.encrypt(str(data).encode()).decode()
    return encrypted_text


def fernet_decrypt(data: str, key: str = None, is_dict=False) -> str:
    """解密文本

    :param data: 需要解密的文本
    :param key: Fernet key must be 32 url-safe base64-encoded bytes.
        推荐使用 generate_fernet_key() 生成
    :param is_dict: 是否解密字典数据
    :return: 解密后的文本
    """
    from cryptography.fernet import Fernet

    key = key or os.getenv("FERNET_KEY")
    cipher_suite = Fernet(key.encode())
    decrypted_text = cipher_suite.decrypt(data.encode()).decode()
    return eval(decrypted_text) if is_dict else decrypted_text


def test():
    key = generate_fernet_key()
    # key = 'HYtUW7y0HOMQySGmOiDHztUGaHC-WnBVh-yqn11Tszw='
    text = {"account": "admin", "password": "123456"}
    encrypted = fernet_encrypt(text, key)
    decrypted = fernet_decrypt(encrypted, key, is_dict=True)
    assert text == decrypted, f"{text} != {decrypted}"



import pandas as pd
from czsc.utils.calendar import is_trading_date, next_trading_date, prev_trading_date


def test_is_trading_date():
    test_cases = [
        ('2023-09-08', True),
        ('2023-09-09', False),
        ('2023-09-10', False),
        ('2023-09-10 12:00', False),
        ('2024-04-18', True),
        ('2024-01-13', False)
    ]
    for date, expected_result in test_cases:
        assert is_trading_date(date) == expected_result


def test_next_trading_date():
    assert next_trading_date('2023-09-09') == pd.Timestamp('2023-09-11')
    assert next_trading_date('2023-09-09', n=2) == pd.Timestamp('2023-09-12')
    assert next_trading_date('2023-09-12', n=1) == pd.Timestamp('2023-09-13')
    assert next_trading_date('2023-09-12 12:00', n=1) == pd.Timestamp('2023-09-13')


def test_prev_trading_date():
    assert prev_trading_date('2023-09-09') == pd.Timestamp('2023-09-08')
    assert prev_trading_date('2023-09-09', n=2) == pd.Timestamp('2023-09-07')
    assert prev_trading_date('2023-09-07 12:00', n=1) == pd.Timestamp('2023-09-06')


def test_get_trading_dates():
    from czsc.utils.calendar import get_trading_dates
    dates = get_trading_dates('2023-09-08', '2023-09-12')
    assert dates == pd.to_datetime(['2023-09-08', '2023-09-11', '2023-09-12']).tolist()
    dates = get_trading_dates('2023-09-08 12:00', '2023-09-12 15:00')
    assert dates == pd.to_datetime(['2023-09-08', '2023-09-11', '2023-09-12']).tolist()



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/2/7 13:17
describe: 用于探索性分析的函数
"""
import loguru
import pandas as pd
import numpy as np
from typing import Callable
from tqdm import tqdm


def vwap(price: np.array, volume: np.array, **kwargs) -> float:
    """计算成交量加权平均价

    :param price: 价格序列
    :param volume: 成交量序列
    :return: 平均价
    """
    return np.average(price, weights=volume)


def twap(price: np.array, **kwargs) -> float:
    """计算时间加权平均价

    :param price: 价格序列
    :return: 平均价
    """
    return np.average(price)


def remove_beta_effects(df, **kwargs):
    """去除 beta 对因子的影响

    :param df: DataFrame, 数据, 必须包含 dt、symbol、factor 和 betas 列
    :param kwargs:

        - factor: str, 因子列名
        - betas: list, beta 列名列表
        - linear_model: str, 线性模型，可选 ridge、linear 或 lasso
        - linear_model_params: dict, 线性模型参数, 默认为空, 需要传入字典，根据模型不同参数不同

    :return: DataFrame
    """
    from sklearn.linear_model import Ridge, LinearRegression, Lasso

    linear_model = kwargs.get("linear_model", "ridge")
    linear_model_params = kwargs.get("linear_model_params", {})
    linear = {
        "ridge": Ridge,
        "linear": LinearRegression,
        "lasso": Lasso,
    }
    assert linear_model in linear.keys(), "linear_model 参数必须为 ridge、linear 或 lasso"
    Model = linear[linear_model]

    factor = kwargs.get("factor")
    betas = kwargs.get("betas")
    logger = kwargs.get("logger", loguru.logger)

    assert factor is not None and betas is not None, "factor 和 betas 参数必须指定"
    assert isinstance(betas, list), "betas 参数必须为列表"
    assert factor in df.columns, f"数据中不包含因子 {factor}"
    assert all([x in df.columns for x in betas]), f"数据中不包含全部 beta {betas}"

    logger.info(f"去除 beta 对因子 {factor} 的影响, 使用 {linear_model} 模型, betas: {betas}")

    rows = []
    for dt, dfg in df.groupby("dt"):
        dfg = dfg.copy().dropna(subset=[factor] + betas)
        if dfg.empty:
            continue

        x = dfg[betas].values
        y = dfg[factor].values
        model = Model(**linear_model_params).fit(x, y)
        dfg[factor] = y - model.predict(x)
        rows.append(dfg)

    dfr = pd.concat(rows, ignore_index=True)
    return dfr


def cross_sectional_strategy(df, factor, weight="weight", long=0.3, short=0.3, **kwargs):
    """根据截面因子值构建多空组合

    :param df: pd.DataFrame, 包含因子列的数据, 必须包含 dt, symbol, factor 列
    :param factor: str, 因子列名称
    :param weight: str, 权重列名称，默认为 weight
    :param long: float, 多头持仓比例/数量，默认为 0.3, 取值范围为 [0, n_symbols], 0~1 表示比例，大于等于1表示数量
    :param short: float, 空头持仓比例/数量，默认为 0.3, 取值范围为 [0, n_symbols], 0~1 表示比例，大于等于1表示数量
    :param kwargs:

        - factor_direction: str, 因子方向，positive 或 negative
        - logger: loguru.logger, 日志记录器
        - norm: bool, 是否对 weight 进行截面持仓标准化，默认为 True

    :return: pd.DataFrame, 包含 weight 列的数据
    """
    factor_direction = kwargs.get("factor_direction", "positive")
    logger = kwargs.get("logger", loguru.logger)
    norm = kwargs.get("norm", True)

    assert long >= 0 and short >= 0, "long 和 short 参数必须大于等于0"
    assert factor in df.columns, f"{factor} 不在 df 中"
    assert factor_direction in ["positive", "negative"], f"factor_direction 参数错误"

    df = df.copy()
    if factor_direction == "negative":
        df[factor] = -df[factor]

    df[weight] = 0.0
    rows = []

    for dt, dfg in df.groupby("dt"):
        long_num = int(long) if long >= 1 else int(len(dfg) * long)
        short_num = int(short) if short >= 1 else int(len(dfg) * short)

        if long_num == 0 and short_num == 0:
            logger.warning(f"{dt} 多空目前持仓数量都为0; long: {long}, short: {short}")
            rows.append(dfg)
            continue

        long_symbols = dfg.sort_values(factor, ascending=False).head(long_num)['symbol'].tolist()
        short_symbols = dfg.sort_values(factor, ascending=True).head(short_num)['symbol'].tolist()

        union_symbols = set(long_symbols) & set(short_symbols)
        if union_symbols:
            logger.warning(f"{dt} 存在同时在多头和空头的品种：{union_symbols}")
            long_symbols = list(set(long_symbols) - union_symbols)
            short_symbols = list(set(short_symbols) - union_symbols)

        dfg.loc[dfg['symbol'].isin(long_symbols), weight] = 1 / long_num if norm else 1.0
        dfg.loc[dfg['symbol'].isin(short_symbols), weight] = -1 / short_num if norm else -1.0
        rows.append(dfg)

    dfx = pd.concat(rows, ignore_index=True)
    return dfx


def judge_factor_direction(df: pd.DataFrame, factor, target='n1b', by='symbol', **kwargs):
    """判断因子的方向，正向还是反向

    :param df: pd.DataFrame, 数据源，必须包含 symbol, dt, target, factor 列
    :param factor: str, 因子名称
    :param target: str, 目标名称，默认为 n1b，表示下一根K线的涨跌幅
    :param by: str, 分组字段，默认为 symbol，表示按品种分组（时序）；也可以按 dt 分组，表示按时间分组（截面）
    :param kwargs: dict, 其他参数
        - method: str, 相关系数计算方法，默认为 pearson，可选 pearson, kendall, spearman
    :return: str, positive or negative
    """
    assert by in df.columns, f"数据中不存在 {by} 字段"
    assert factor in df.columns, f"数据中不存在 {factor} 字段"
    assert target in df.columns, f"数据中不存在 {target} 字段"

    if by == "dt" and df['symbol'].nunique() < 2:
        raise ValueError("品种数量过少，无法在时间截面上计算因子有效性方向")

    if by == "symbol" and df['dt'].nunique() < 2:
        raise ValueError("时间序列数据量过少，无法在品种上计算因子有效性方向")

    method = kwargs.get("method", "pearson")
    dfc = df.groupby(by)[[factor, target]].corr(method=method).unstack().iloc[:, 1].reset_index()
    return "positive" if dfc[factor].mean().iloc[0] >= 0 else "negative"


def monotonicity(sequence):
    """计算序列的单调性

    原理：计算序列与自然数序列的相关系数，系数越接近1，表示单调递增；系数越接近-1，表示单调递减；接近0表示无序

    :param sequence: list, tuple 序列
    :return: float, 单调性系数
    """
    from scipy.stats import spearmanr
    return spearmanr(sequence, range(len(sequence)))[0]


def min_max_limit(x, min_val, max_val, digits=4):
    """限制 x 的取值范围在 min_val 和 max_val 之间

    :param x: float, 输入值
    :param min_val: float, 最小值
    :param max_val: float, 最大值
    :param digits: int, 保留小数位数
    :return: float
    """
    return round(max(min_val, min(max_val, x)), digits)


def rolling_layers(df, factor, n=5, **kwargs):
    """对时间序列数据进行分层

    :param df: 因子数据，必须包含 dt, factor 列，其中 dt 为日期，factor 为因子值
    :param factor: 因子列名
    :param n: 分层数量，默认为10
    :param kwargs:

        - window: 窗口大小，默认为2000
        - min_periods: 最小样本数量，默认为300
        - mode: str, {'loose', 'strict'}, 分层模式，默认为 'loose'；
            loose 表示使用 rolling + rank 的方式分层，有一点点未来信息，存在一定的数据穿越问题；
            strict 表示使用 rolling + qcut 的方式分层，无未来信息，但是执行速度较慢。

    :return: df, 添加了 factor分层 列
    """
    assert df[factor].nunique() > n * 2, "因子值的取值数量必须大于分层数量"
    assert df[factor].isna().sum() == 0, "因子有缺失值，缺失数量为：{}".format(df[factor].isna().sum())
    assert df['dt'].duplicated().sum() == 0, f"dt 列不能有重复值，存在重复值数量：{df['dt'].duplicated().sum()}"

    window = kwargs.get("window", 600)
    min_periods = kwargs.get("min_periods", 300)

    # 不能有 inf 和 -inf
    if df.loc[df[factor].isin([float("inf"), float("-inf")]), factor].shape[0] > 0:
        raise ValueError(f"存在 {factor} 为 inf / -inf 的数据")

    if kwargs.get('mode', 'loose') == 'loose':
        # loose 模式，可能存在一点点未来信息
        df['pct_rank'] = df[factor].rolling(window=window, min_periods=min_periods).rank(pct=True, ascending=True)
        bins = [i/n for i in range(n+1)]
        df['pct_rank_cut'] = pd.cut(df['pct_rank'], bins=bins, labels=False)
        df['pct_rank_cut'] = df['pct_rank_cut'].fillna(-1)
        # 第00层表示缺失值
        df[f"{factor}分层"] = df['pct_rank_cut'].apply(lambda x: f"第{str(int(x+1)).zfill(2)}层")
        df.drop(['pct_rank', 'pct_rank_cut'], axis=1, inplace=True)

    else:
        assert kwargs.get('mode', 'strict') == 'strict'
        df[f"{factor}_qcut"] = (
            df[factor].rolling(window=window, min_periods=min_periods)
            .apply(lambda x: pd.qcut(x, q=n, labels=False, duplicates="drop", retbins=False).values[-1], raw=False)
        )
        df[f"{factor}_qcut"] = df[f"{factor}_qcut"].fillna(-1)
        # 第00层表示缺失值
        df[f"{factor}分层"] = df[f"{factor}_qcut"].apply(lambda x: f"第{str(int(x+1)).zfill(2)}层")
        df.drop([f"{factor}_qcut"], axis=1, inplace=True)

    return df


def cal_yearly_days(dts: list, **kwargs):
    """计算年度交易日数量

    :param dts: list, datetime 列表
    :param kwargs:
    :return: int, 年度交易日数量
    """
    logger = kwargs.get("logger", loguru.logger)

    assert len(dts) > 0, "输入的日期数量必须大于0"

    # 将日期列表转换为 DataFrame
    dts = pd.DataFrame(dts, columns=["dt"])
    dts["dt"] = pd.to_datetime(dts["dt"]).dt.date
    dts = dts.drop_duplicates()

    # 时间跨度小于一年，直接返回252，并警告
    if (dts["dt"].max() - dts["dt"].min()).days < 365:
        logger.warning("时间跨度小于一年，直接返回 252")
        return 252

    # 设置索引为日期，并确保索引为 DatetimeIndex
    dts.set_index(pd.to_datetime(dts["dt"]), inplace=True)
    dts.drop(columns=["dt"], inplace=True)

    # 按年重采样并计算每年的交易日数量，取最大值
    yearly_days = dts.resample('YE').size().max()
    return min(yearly_days, 365)


def cal_symbols_factor(dfk: pd.DataFrame, factor_function: Callable, **kwargs):
    """计算多个品种的标准量价因子

    :param dfk: 行情数据，N 个品种的行情数据
    :param factor_function: 因子文件，py文件
    :param kwargs:

        - logger: loguru.logger, 默认为 loguru.logger
        - factor_params: dict, 因子计算参数
        - min_klines: int, 最小K线数据量，默认为 300
        - price_type: str, 交易价格类型，默认为 close，可选值为 close 或 next_open
        - strict: bool, 是否严格模式，默认为 True, 严格模式下，计算因子出错会抛出异常

    :return: dff, pd.DataFrame, 计算后的因子数据
    """
    logger = kwargs.get("logger", loguru.logger)
    min_klines = kwargs.get("min_klines", 300)
    factor_params = kwargs.get("factor_params", {})
    price_type = kwargs.get("price_type", "close")
    strict = kwargs.get("strict", True)

    symbols = dfk["symbol"].unique().tolist()
    factor_name = factor_function.__name__

    def __one_symbol(symbol):
        df = dfk[(dfk["symbol"] == symbol)].copy()
        df = df.sort_values("dt", ascending=True).reset_index(drop=True)
        if len(df) < min_klines:
            logger.warning(f"{symbol} 数据量过小，跳过；仅有 {len(df)} 条数据，需要 {min_klines} 条数据")
            return None

        df = factor_function(df, **factor_params)
        if price_type == 'next_open':
            df["price"] = df["open"].shift(-1).fillna(df["close"])
        elif price_type == 'close':
            df["price"] = df["close"]
        else:
            raise ValueError("price_type 参数错误, 可选值为 close 或 next_open")

        df["n1b"] = (df["price"].shift(-1) / df["price"] - 1).fillna(0)
        factor = [x for x in df.columns if x.startswith("F#")][0]

        # df[factor] = df[factor].replace([np.inf, -np.inf], np.nan).ffill().fillna(0)
        # factor 中不能有 inf 和 -inf 值，也不能有 nan 值
        assert df[factor].isna().sum() == 0, f"{symbol} {factor} 存在 nan 值"
        assert df[factor].isin([np.inf, -np.inf]).sum() == 0, f"{symbol} {factor} 存在 inf 值"
        assert df[factor].var() != 0 and not np.isnan(df[factor].var()), f"{symbol} {factor} var is 0 or nan"
        return df

    rows = []
    for _symbol in tqdm(symbols, desc=f"{factor_name} 因子计算"):
        if strict:
            dfx = __one_symbol(_symbol)
        else:
            try:
                dfx = __one_symbol(_symbol)
            except Exception as e:
                logger.error(f"{factor_name} - {_symbol} - 计算因子出错：{e}")
                continue
        rows.append(dfx)

    dff = pd.concat(rows, ignore_index=True)
    return dff


def weights_simple_ensemble(df, weight_cols, method="mean", only_long=False, **kwargs):
    """用朴素的方法集成多个策略的权重

    :param df: pd.DataFrame, 包含多个策略的权重列
    :param weight_cols: list, 权重列名称列表
    :param method: str, 集成方法，可选 mean, vote, sum_clip

        - mean: 平均值
        - vote: 投票
        - sum_clip: 求和并截断

    :param only_long: bool, 是否只做多
    :param kwargs: dict, 其他参数

        - clip_min: float, 截断最小值
        - clip_max: float, 截断最大值

    :return: pd.DataFrame, 添加了 weight 列的数据
    """
    method = method.lower()

    assert all([x in df.columns for x in weight_cols]), f"数据中不包含全部权重列，不包含的列：{set(weight_cols) - set(df.columns)}"
    assert 'weight' not in df.columns, "数据中已经包含 weight 列，请先删除，再调用该函数"

    if method == "mean":
        df["weight"] = df[weight_cols].mean(axis=1).fillna(0)

    elif method == "vote":
        df["weight"] = np.sign(df[weight_cols].sum(axis=1)).fillna(0)

    elif method == "sum_clip":
        clip_min = kwargs.get("clip_min", -1)
        clip_max = kwargs.get("clip_max", 1)
        df["weight"] = df[weight_cols].sum(axis=1).clip(clip_min, clip_max).fillna(0)

    else:
        raise ValueError("method 参数错误，可选 mean, vote, sum_clip")

    if only_long:
        df["weight"] = np.where(df["weight"] > 0, df["weight"], 0)

    return df


def unify_weights(dfw: pd.DataFrame, **kwargs):
    """按策略统一权重进行大盘择时交易

    在任意时刻 dt，将所有品种的权重通过某种算法合并，然后所有品种都按照这个权重进行操作

    :param dfw: pd.DataFrame，columns=['symbol', 'weight', 'dt', 'price']，数据样例如下

        ========  ===================  ========  =======
        symbol    dt                     weight    price
        ========  ===================  ========  =======
        IC9001    2017-01-03 00:00:00     -0.82  11113.8
        IC9001    2017-01-04 00:00:00     -0.83  11275.3
        IC9001    2017-01-05 00:00:00     -0.84  11261.1
        ========  ===================  ========  =======

    :param kwargs: dict，其他参数

        - method: str，权重合并方法，支持 'mean' 和 'sum_clip'，默认 'sum_clip'
        - copy: bool，是否复制输入数据，默认 True
        - clip_min: float，权重合并方法为 'sum_clip' 时，clip 的最小值，默认 -1
        - clip_max: float，权重合并方法为 'sum_clip' 时，clip 的最大值，默认 1

    :return: pd.DataFrame，columns=['symbol', 'weight', 'dt', 'price']

    example:
    ================
        dfw = ...
        wb = czsc.WeightBacktest(dfw, fee_rate=0.0002)
        print(wb.stats)

        dfw1 = unify_weights(dfw.copy(), method='mean')
        wb1 = czsc.WeightBacktest(dfw1, fee_rate=0.0002)
        print(wb1.stats)

        dfw2 = unify_weights(dfw.copy(), method='sum_clip')
        wb2 = czsc.WeightBacktest(dfw2, fee_rate=0.0002)
        print(wb2.stats)

        # 合并 daily_return，看看是否一致
        dfd1 = wb.daily_return.copy()
        dfd2 = wb1.daily_return.copy()
        dfd3 = wb2.daily_return.copy()

        dfd = pd.merge(dfd1, dfd2, on='date', how='left', suffixes=('', '_mean'))
        dfd = pd.merge(dfd, dfd3, on='date', how='left', suffixes=('', '_sum_clip'))
        print(dfd[['total', 'total_mean', 'total_sum_clip']].corr())
    ================
    """
    method = kwargs.get('method', 'sum_clip')
    if kwargs.get("copy", True):
        dfw = dfw.copy()

    if method == 'mean':
        uw = dfw.groupby('dt')['weight'].mean().reset_index()

    elif method == 'sum_clip':
        clip_min = kwargs.get('clip_min', -1)
        clip_max = kwargs.get('clip_max', 1)
        assert clip_min < clip_max, "clip_min should be less than clip_max"

        uw = dfw.groupby('dt')['weight'].sum().reset_index()
        uw['weight'] = uw['weight'].clip(clip_min, clip_max)

    else:
        raise ValueError(f"method {method} not supported")

    dfw = pd.merge(dfw, uw, on='dt', how='left', suffixes=('_raw', '_unified'))
    dfw['weight'] = dfw['weight_unified'].copy()
    return dfw


def sma_long_bear(df: pd.DataFrame, **kwargs):
    """均线牛熊分界指标过滤持仓，close 在长期均线上方为牛市，下方为熊市

    牛市只做多，熊市只做空。

    :param df: DataFrame, 必须包含 dt, close, symbol, weight 列
    :return: DataFrame
    """
    assert df["symbol"].nunique() == 1, "数据中包含多个品种，必须单品种"
    assert df["dt"].is_monotonic_increasing, "数据未按日期排序，必须升序排列"
    assert df["dt"].is_unique, "数据中存在重复dt，必须唯一"

    window = kwargs.get("window", 200)

    if kwargs.get("copy", True):
        df = df.copy()

    df['SMA_LB'] = df['close'].rolling(window).mean()
    df['raw_weight'] = df['weight']
    df['weight'] = np.where(np.sign(df['close'] - df['SMA_LB']) == np.sign(df['weight']), df['weight'], 0)
    return df


def dif_long_bear(df: pd.DataFrame, **kwargs):
    """DIF牛熊分界指标过滤持仓，DIF 在0上方为牛市，下方为熊市

    牛市只做多，熊市只做空。

    :param df: DataFrame, 必须包含 dt, close, symbol, weight 列
    :return: DataFrame
    """
    from czsc.utils.ta import MACD

    assert df["symbol"].nunique() == 1, "数据中包含多个品种，必须单品种"
    assert df["dt"].is_monotonic_increasing, "数据未按日期排序，必须升序排列"
    assert df["dt"].is_unique, "数据中存在重复dt，必须唯一"

    if kwargs.get("copy", True):
        df = df.copy()

    df['DIF_LB'], _, _ = MACD(df['close'])
    df['raw_weight'] = df['weight']
    df['weight'] = np.where(np.sign(df['DIF_LB']) == np.sign(df['weight']), df['weight'], 0)
    return df


def tsf_type(df: pd.DataFrame, factor, n=5, **kwargs):
    """时序因子的类型定性分析

    tsf 是 time series factor 的缩写，时序因子的类型定性分析，是指对某个时序因子进行分层，然后计算每个分层的平均收益，

    :param df: pd.DataFrame, 必须包含 dt, symbol, factor 列，其中 dt 为日期，symbol 为标的代码，factor 为因子值
    :param factor: str, 因子列名
    :param n: int, 分层数量
    :param kwargs:

        - window: int, 窗口大小，默认为600
        - min_periods: int, 最小样本数量，默认为300
        - target: str, 目标列名，默认为 n1b

    :return: str, 返回分层收益排序（从大到小）结果，例如：第01层->第02层->第03层->第04层->第05层
    """
    logger = kwargs.get("logger", loguru.logger)
    window = kwargs.get("window", 600)
    min_periods = kwargs.get("min_periods", 300)
    target = kwargs.get("target", "n1b")

    if target == 'n1b' and 'n1b' not in df.columns:
        from czsc.utils.trade import update_nxb
        df = update_nxb(df, nseq=(1,))

    assert target in df.columns, f"数据中不存在 {target} 列"
    assert factor in df.columns, f"数据中不存在 {factor} 列"

    rows = []
    for symbol, dfg in df.groupby("symbol"):
        try:
            dfg = dfg.copy().reset_index(drop=True)
            dfg = rolling_layers(dfg, factor, n=n, window=window, min_periods=min_periods)
            rows.append(dfg)
        except Exception as e:
            logger.warning(f"{symbol} 计算分层失败: {e}")

    df = pd.concat(rows, ignore_index=True)
    layers = [x for x in df[f"{factor}分层"].unique() if x != "第00层" and str(x).endswith("层")]

    # 计算每个分层的平均收益
    layer_returns = {}
    for layer in layers:
        dfg = df[df[f"{factor}分层"] == layer].copy()
        dfg = dfg.groupby("dt")[target].mean().reset_index()
        layer_returns[layer] = dfg[target].sum()

    sorted_layers = sorted(layer_returns.items(), key=lambda x: x[1], reverse=True)
    return "->".join([f"{x[0]}" for x in sorted_layers])


def limit_leverage(df: pd.DataFrame, leverage: float = 1.0, **kwargs):
    """限制杠杆比例

    原理描述：

    1. 计算滚动窗口内权重的绝对均值 abs_mean，初始窗口内权重的绝对均值设为 leverage
    2. 用 leverage 除以 abs_mean，得到调整比例 adjust_ratio
    3. 将原始权重乘以 adjust_ratio，再限制在 -leverage 和 leverage 之间

    :param df: DataFrame, columns=['dt', 'symbol', 'weight']
    :param leverage: float, 杠杆倍数
    :param kwargs:

        - copy: bool, 是否复制 DataFrame
        - window: int, 滚动窗口，默认为 300
        - min_periods: int, 最小样本数，小于该值的窗口不计算均值，默认为 50
        - weight: str, 权重列名，默认为 'weight'
        - method: str, 计算均值的方法，'abs_mean' 或 'abs_max'，默认为 'abs_mean'
            abs_mean: 计算绝对均值作为调整杠杆的标准
            abs_max: 计算绝对最大值作为调整杠杆的标准

    :return: DataFrame
    """
    window = kwargs.get("window", 300)
    min_periods = kwargs.get("min_periods", 50)
    weight = kwargs.get("weight", "weight")
    method = kwargs.get("method", "abs_mean")

    assert weight in df.columns, f"数据中不包含权重列 {weight}"

    if kwargs.get("copy", False):
        df = df.copy()

    df = df.sort_values(["dt", "symbol"], ascending=True).reset_index(drop=True)

    for symbol in df['symbol'].unique():
        dfx = df[df['symbol'] == symbol].copy()
        # assert dfx['dt'].is_monotonic_increasing, f"{symbol} 数据未按日期排序，必须升序排列"
        assert dfx['dt'].is_unique, f"{symbol} 数据中存在重复dt，必须唯一"

        if method == "abs_mean":
            bench = dfx[weight].abs().rolling(window=window, min_periods=min_periods).mean().fillna(leverage)
        elif method == "abs_max":
            bench = dfx[weight].abs().rolling(window=window, min_periods=min_periods).max().fillna(leverage)
        else:
            raise ValueError(f"不支持的 method: {method}")

        adjust_ratio = leverage / bench
        df.loc[df['symbol'] == symbol, weight] = (dfx[weight] * adjust_ratio).clip(-leverage, leverage)

    return df




# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/5/11 17:53
describe:
"""
import pandas as pd


def cross_sectional_ranker(df, x_cols, y_col, **kwargs):
    """截面打分排序

    :param df: 因子数据，必须包含日期、品种、因子值、预测列，且按日期升序排列，样例数据如下：
    :param x_cols: 因子列名
    :param y_col: 预测列名
    :param kwargs: 其他参数

        - model_params: dict, 模型参数，默认{'n_estimators': 40, 'learning_rate': 0.01}，可调整，参考lightgbm文档
        - n_splits: int, 时间拆分次数，默认5，即5段时间
        - rank_ascending: bool, 打分排序是否升序，默认False-降序
        - copy: bool, 是否拷贝df，True-拷贝，False-不拷贝

    :return: df, 包含预测分数和排序列
    """
    from lightgbm import LGBMRanker
    from sklearn.model_selection import TimeSeriesSplit

    assert "symbol" in df.columns, "df must have column 'symbol'"
    assert "dt" in df.columns, "df must have column 'dt'"

    if kwargs.get("copy", True):
        df = df.copy()
    df["dt"] = pd.to_datetime(df["dt"])
    df = df.sort_values(["dt", y_col], ascending=[True, False])

    model_params = kwargs.get("model_params", {"n_estimators": 40, "learning_rate": 0.01})
    model = LGBMRanker(**model_params)

    dfd = pd.DataFrame({"dt": sorted(df["dt"].unique())}).values
    tss = TimeSeriesSplit(n_splits=kwargs.get("n_splits", 5))

    for train_index, test_index in tss.split(dfd):
        train_dts = dfd[train_index][:, 0]
        test_dts = dfd[test_index][:, 0]

        # 拆分训练集和测试集
        train, test = df[df["dt"].isin(train_dts)], df[df["dt"].isin(test_dts)]
        X_train, X_test, y_train = train[x_cols], test[x_cols], train[y_col]
        query_train = train.groupby("dt")["symbol"].count().values

        # 训练模型 & 预测
        model.fit(X_train, y_train, group=query_train)
        df.loc[X_test.index, "score"] = model.predict(X_test)

    df["rank"] = df.groupby("dt")["score"].rank(ascending=kwargs.get("rank_ascending", False))
    return df



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/3/23 19:12
describe:
"""
import os
import time
import pandas as pd
from tqdm import tqdm
from loguru import logger
from concurrent.futures import ProcessPoolExecutor
from czsc import fsa
from czsc.traders.base import generate_czsc_signals


class DummyBacktest:
    def __init__(self, strategy, signals_path, results_path, read_bars, **kwargs):
        """策略回测（支持多进程执行）

        :param strategy: CZSC择时策略
        :param signals_path: 信号文件存放路径
        :param results_path: 回测结果存放路径
        :param read_bars: 读入K线数据的函数
            函数签名为：read_bars(symbol, freq, sdt, edt, fq) -> List[RawBar]
        :param kwargs: 其他参数
            - signals_module_name: 信号函数模块名，用于动态加载信号文件，默认为 czsc.signals
        """
        from czsc.strategies import CzscStrategyBase
        assert issubclass(strategy, CzscStrategyBase), "strategy 必须是 CzscStrategyBase 的子类"
        self.strategy = strategy
        self.results_path = results_path
        os.makedirs(self.results_path, exist_ok=True)
        self.signals_path = signals_path
        os.makedirs(self.signals_path, exist_ok=True)
        # 缓存 poss 数据
        self.poss_path = os.path.join(results_path, 'poss')
        os.makedirs(self.poss_path, exist_ok=True)
        logger.add(os.path.join(self.results_path, 'dummy.log'), encoding='utf-8', enqueue=True)
        self.read_bars = read_bars
        self.kwargs = kwargs

        # 回测起止时间
        self.sdt = kwargs.get('sdt', '20100101')
        self.edt = kwargs.get('edt', '20230301')
        self.bars_sdt = pd.to_datetime(self.sdt) - pd.Timedelta(days=365*3)

    def replay(self, symbol):
        """回放单个品种的交易"""
        tactic = self.strategy(symbol=symbol, **self.kwargs)
        bars = self.read_bars(symbol, tactic.base_freq, self.sdt, self.edt, fq='后复权')
        tactic.replay(bars, os.path.join(self.results_path, f"{symbol}_replay"), sdt='20200101')

    def one_symbol_dummy(self, symbol):
        """回测单个品种"""
        start_time = time.time()
        tactic = self.strategy(symbol=symbol, **self.kwargs)
        symbol_path = os.path.join(self.poss_path, symbol)
        if os.path.exists(symbol_path):
            logger.info(f"{symbol} 已经回测过，跳过")
            return None

        os.makedirs(symbol_path, exist_ok=True)
        try:
            file_sigs = os.path.join(self.signals_path, f"{symbol}.sigs")
            if not os.path.exists(file_sigs):
                bars = self.read_bars(symbol, tactic.base_freq, self.bars_sdt, self.edt, fq='后复权')
                sigs = generate_czsc_signals(bars, signals_config=tactic.signals_config, sdt=self.sdt, df=True)
                sigs.drop(columns=['freq', 'cache'], inplace=True)
                sigs.to_parquet(file_sigs)
            else:
                sigs = pd.read_parquet(file_sigs)
                sigs = sigs[sigs['dt'] >= self.sdt]

            sigs = sigs.to_dict('records')
            trader = tactic.dummy(sigs)

        except Exception as e:
            logger.exception(e)
            return None

        for pos in trader.positions:
            try:
                file_pairs = os.path.join(symbol_path, f"{pos.name}.pairs")
                file_holds = os.path.join(symbol_path, f"{pos.name}.holds")

                pairs = pd.DataFrame(pos.pairs)
                pairs.to_parquet(file_pairs)

                dfh = pd.DataFrame(pos.holds)
                dfh['n1b'] = (dfh['price'].shift(-1) / dfh['price'] - 1) * 10000
                dfh.fillna(0, inplace=True)
                dfh['symbol'] = pos.symbol
                dfh.to_parquet(file_holds)
            except Exception as e:
                logger.debug(f"{symbol} {pos.name} 保存失败，原因：{e}")

        logger.info(f"{symbol} 回测完成，共 {len(trader.positions)} 个持仓策略，耗时 {time.time() - start_time:.2f} 秒")

    def one_pos_stats(self, pos_name):
        """分析单个持仓策略的表现"""
        from czsc.traders.performance import PairsPerformance

        symbols = os.listdir(self.poss_path)
        pos_pairs = []
        pos_holds = []
        for symbol in tqdm(symbols, desc=f"读取 {pos_name}"):
            try:
                dfp = pd.read_parquet(os.path.join(self.poss_path, f"{symbol}/{pos_name}.pairs"))
                pos_pairs.append(dfp)

                dfh = pd.read_parquet(os.path.join(self.poss_path, f"{symbol}/{pos_name}.holds"))
                pos_holds.append(dfh[dfh['pos'] != 0])
            except Exception as e:
                logger.debug(f"{symbol} 读取失败，原因：{e}")

        pairs = pd.concat(pos_pairs, ignore_index=True)
        if not pairs.empty:
            pp = PairsPerformance(pairs)
            pairs.to_feather(os.path.join(self.results_path, f"{pos_name}_pairs.feather"))
            pp.agg_to_excel(os.path.join(self.results_path, f"{pos_name}_回测结果.xlsx"))

            stats = dict(pp.basic_info)
            # 加入截面等权评价
            holds = pd.concat(pos_holds, ignore_index=True)
            cross = holds.groupby('dt').apply(lambda x: (x['n1b'] * x['pos']).sum() / sum(x['pos'] != 0))
            stats['截面等权收益'] = cross.sum()
            cross.to_excel(os.path.join(self.results_path, f"{pos_name}_截面等权收益.xlsx"), index=True)
            stats['pos_name'] = pos_name
            return stats
        else:
            return None

    def execute(self, symbols, n_jobs=2, **kwargs):
        """回测多个品种

        :param symbols: 品种列表
        :param n_jobs: 进程数量，默认为 2
            需要注意的是：
            1. 如果进程数过多，可能会导致内存不足
            2. 多进程在 pycharm 的 ipython 中无法使用，需要在命令行中运行
        :param kwargs:
        :return:
        """
        results_path = self.results_path
        tactic = self.strategy(symbol="symbol", **self.kwargs)
        dumps_map = {pos.name: pos.dump() for pos in tactic.positions}

        logger.info(f"策略回测，持仓策略数量：{len(tactic.positions)}，共 {len(symbols)} 只标的，使用 {n_jobs} 个进程；"
                    f"结果保存在 {results_path}。请耐心等待...")

        with ProcessPoolExecutor(n_jobs) as pool:
            pool.map(self.one_symbol_dummy, sorted(symbols))

        all_stats = []
        with ProcessPoolExecutor(max_workers=min(n_jobs, 6)) as pool:
            _stats = pool.map(self.one_pos_stats, list(dumps_map.keys()))
            for _s in _stats:
                if not _s:
                    continue
                _s['pos_dump'] = dumps_map[_s['pos_name']]
                all_stats.append(_s)

        file_report = os.path.join(results_path, f'{self.strategy.__name__}_回测结果汇总.xlsx')
        report_df = pd.DataFrame(all_stats).sort_values(['截面等权收益'], ascending=False, ignore_index=True)
        report_df.to_excel(file_report, index=False)
        logger.info(f"策略回测完成，结果保存在 {results_path}。")

        if kwargs.get('feishu_app_id') and kwargs.get('feishu_app_secret'):
            if os.path.exists(file_report):
                fsa.push_message(file_report, msg_type='file', **kwargs)
            else:
                fsa.push_message(f"{self.strategy.__name__} 回测结果为空，请检查原因！", **kwargs)



# coding: utf-8
import numpy as np
from collections import OrderedDict
from czsc.utils import x_round
from czsc.objects import Signal, Factor, Event, Freq, Operate
from czsc.objects import cal_break_even_point


# def test_raw_bar():
#     from test.test_analyze import read_daily
#     from czsc.utils.ta import SMA
#     bars = read_daily()
#     ma = SMA(np.array([x.close for x in bars]), 5)
#     key = "SMA5"
#
#     # 技术指标的全部更新
#     for i in range(1, len(bars) + 1):
#         c = dict(bars[-i].cache) if bars[-i].cache else dict()
#         c.update({key: ma[-i]})
#         bars[-i].cache = c
#     assert np.array([x.cache[key] for x in bars]).sum() == ma.sum()
#
#     # 技术指标的部分更新
#     for i in range(1, 101):
#         c = dict(bars[-i].cache) if bars[-i].cache else dict()
#         c.update({key: ma[-i] + 2})
#         bars[-i].cache = c
#     assert np.array([x.cache[key] for x in bars]).sum() == ma.sum() + 200


def test_zs():
    """测试中枢对象"""
    from test.test_analyze import read_daily
    from czsc.objects import ZS
    from czsc.analyze import CZSC

    bars = read_daily()
    c = CZSC(bars)

    zs = ZS(c.bi_list[-5:])
    assert zs.zd < zs.zg
    assert zs.is_valid

    zs = ZS(c.bi_list[-8:-3])
    assert not zs.is_valid


def test_cal_break_even_point():
    assert cal_break_even_point([]) == 1
    assert cal_break_even_point([1]) == 1
    assert cal_break_even_point([-1, -2, 4, 5, 5]) == 0.6
    assert cal_break_even_point([-1, -2]) == 1
    assert cal_break_even_point([1, 2]) == 0.5
    assert cal_break_even_point([-1, 1, -2]) == 1
    assert cal_break_even_point([0, 1, -1, 2, 3, -6, -7]) == 1
    assert x_round(cal_break_even_point([0, 1, -1, 2, 3, -6, 7, 8])) == 0.875
    assert x_round(cal_break_even_point([-6, -1, 0, 1, 2, 3, 7, 8])) == 0.875
    assert x_round(cal_break_even_point([2, 3, 4, 2, 1, 4, 0, 1, -1, 2, 3, -6, 7, 8])) == 0.5714


def test_signal():
    s = Signal(k1="1分钟", k3="倒1形态", v1="类一买", v2="七笔", v3="基础型", score=3)
    assert str(s) == "Signal('1分钟_任意_倒1形态_类一买_七笔_基础型_3')"
    assert s.key == "1分钟_倒1形态"
    s1 = Signal(signal="1分钟_任意_倒1形态_类一买_七笔_基础型_3")
    assert s == s1
    assert s.is_match({"1分钟_倒1形态": "类一买_七笔_基础型_3"})
    assert not s.is_match({"1分钟_倒1形态": "类一买_七笔_特例一_3"})
    assert not s.is_match({"1分钟_倒1形态": "类一买_九笔_基础型_3"})

    s = Signal(k1="1分钟", k2="倒1形态", k3="类一买", score=3)
    assert str(s) == "Signal('1分钟_倒1形态_类一买_任意_任意_任意_3')"
    assert s.key == "1分钟_倒1形态_类一买"

    try:
        s = Signal(k1="1分钟", k2="倒1形态", k3="类一买", score=101)
    except ValueError as e:
        assert str(e) == "score 必须在0~100之间"


def test_factor():
    freq = Freq.F15
    s = OrderedDict()
    default_signals = [
        Signal(k1=str(freq.value), k2="倒0笔", k3="方向", v1="向上", v2="其他", v3="其他"),
        Signal(k1=str(freq.value), k2="倒0笔", k3="长度", v1="大于5", v2="其他", v3="其他"),
        Signal(k1=str(freq.value), k2="倒0笔", k3="三K形态", v1="顶分型", v2="其他", v3="其他"),
        Signal(k1=str(freq.value), k2="倒1笔", k3="表里关系", v1="其他", v2="其他", v3="其他"),
        Signal(k1=str(freq.value), k2="倒1笔", k3="RSQ状态", v1="小于0.2", v2="其他", v3="其他"),
    ]
    for signal in default_signals:
        s[signal.key] = signal.value

    factor = Factor(
        name="单测",
        signals_all=[
            Signal(k1=str(freq.value), k2="倒0笔", k3="方向", v1="向上", v2="其他", v3="其他"),
            Signal(k1=str(freq.value), k2="倒0笔", k3="长度", v1="大于5", v2="其他", v3="其他"),
        ],
    )
    assert factor.is_match(s)

    factor_raw = factor.dump()
    new_factor = Factor.load(factor_raw)
    assert new_factor.is_match(s)

    factor = Factor(
        name="单测",
        signals_all=[
            Signal(k1=str(freq.value), k2="倒0笔", k3="方向", v1="向上", v2="其他", v3="其他"),
            Signal(k1=str(freq.value), k2="倒0笔", k3="长度", v1="大于5", v2="其他", v3="其他"),
        ],
        signals_any=[Signal(k1=str(freq.value), k2="倒1笔", k3="RSQ状态", v1="小于0.2", v2="其他", v3="其他")],
    )
    assert factor.is_match(s)

    factor = Factor(
        name="单测",
        signals_all=[
            Signal(k1=str(freq.value), k2="倒0笔", k3="方向", v1="向上", v2="其他", v3="其他"),
            Signal(k1=str(freq.value), k2="倒0笔", k3="长度", v1="大于5", v2="其他", v3="其他"),
        ],
        signals_any=[Signal(k1=str(freq.value), k2="倒1笔", k3="RSQ状态", v1="小于0.8", v2="其他", v3="其他")],
    )
    assert not factor.is_match(s)

    factor = Factor(
        name="单测",
        signals_all=[
            Signal(k1=str(freq.value), k2="倒0笔", k3="方向", v1="向上", v2="其他", v3="其他"),
            Signal(k1=str(freq.value), k2="倒0笔", k3="长度", v1="大于5", v2="其他", v3="其他"),
        ],
        signals_any=[Signal(k1=str(freq.value), k2="倒1笔", k3="RSQ状态", v1="小于0.2", v2="其他", v3="其他")],
        signals_not=[
            Signal(k1=str(freq.value), k2="倒0笔", k3="三K形态", v1="顶分型", v2="其他", v3="其他"),
        ],
    )
    assert not factor.is_match(s)


def test_event():
    freq = Freq.F15
    s = OrderedDict()
    default_signals = [
        Signal(k1=str(freq.value), k2="倒0笔", k3="方向", v1="向上", v2="其他", v3="其他"),
        Signal(k1=str(freq.value), k2="倒0笔", k3="长度", v1="大于5", v2="其他", v3="其他"),
        Signal(k1=str(freq.value), k2="倒0笔", k3="三K形态", v1="顶分型", v2="其他", v3="其他"),
        Signal(k1=str(freq.value), k2="倒1笔", k3="表里关系", v1="其他", v2="其他", v3="其他"),
        Signal(k1=str(freq.value), k2="倒1笔", k3="RSQ状态", v1="小于0.2", v2="其他", v3="其他"),
    ]
    for signal in default_signals:
        s[signal.key] = signal.value

    event = Event(
        name="单测",
        operate=Operate.LO,
        factors=[
            Factor(
                name="测试",
                signals_all=[Signal(k1=str(freq.value), k2="倒0笔", k3="长度", v1="大于5", v2="其他", v3="其他")],
            )
        ],
        signals_all=[
            Signal(k1=str(freq.value), k2="倒0笔", k3="方向", v1="向上", v2="其他", v3="其他"),
        ],
    )
    m, f = event.is_match(s)
    assert m and f

    raw = event.dump()
    new_event = Event.load(raw)
    m, f = new_event.is_match(s)
    assert m and f

    raw1 = {
        "name": "单测",
        "operate": "开多",
        "signals_all": ["15分钟_倒0笔_方向_向上_其他_其他_0"],
        "factors": [{"name": "测试", "signals_all": ["15分钟_倒0笔_长度_大于5_其他_其他_0"]}],
    }
    new_event = Event.load(raw1)
    m, f = new_event.is_match(s)
    assert m and f

    raw1 = {
        "operate": "开多",
        "signals_all": ["15分钟_倒0笔_方向_向上_其他_其他_0"],
        "factors": [{"name": "测试", "signals_all": ["15分钟_倒0笔_长度_大于5_其他_其他_0"]}],
    }
    new_event = Event.load(raw1)
    m, f = new_event.is_match(s)
    assert m and f

    event = Event(
        name="单测",
        operate=Operate.LO,
        factors=[
            Factor(name="测试", signals_all=[Signal("15分钟_倒0笔_长度_大于5_其他_其他_0")]),
        ],
        signals_any=[Signal("15分钟_倒0笔_方向_向上_其他_其他_0"), Signal("15分钟_倒0笔_长度_大于100_其他_其他_0")],
    )
    m, f = event.is_match(s)
    assert m and f

    event = Event(
        name="单测",
        operate=Operate.LO,
        factors=[
            Factor(
                name="测试",
                signals_all=[Signal(k1=str(freq.value), k2="倒0笔", k3="长度", v1="大于5", v2="其他", v3="其他")],
            )
        ],
        signals_not=[
            Signal(k1=str(freq.value), k2="倒0笔", k3="方向", v1="向上", v2="其他", v3="其他"),
        ],
    )
    m, f = event.is_match(s)
    assert not m and not f

    event = Event(
        name="单测",
        operate=Operate.LO,
        factors=[
            Factor(
                name="测试",
                signals_all=[
                    Signal(k1=str(freq.value), k2="倒0笔", k3="方向", v1="向上", v2="其他", v3="其他"),
                    Signal(k1=str(freq.value), k2="倒0笔", k3="长度", v1="大于5", v2="其他", v3="其他"),
                ],
            )
        ],
    )
    m, f = event.is_match(s)
    assert m and f

    event = Event(
        name="单测",
        operate=Operate.LO,
        factors=[
            Factor(
                name="测试",
                signals_all=[
                    Signal("15分钟_倒0笔_方向_向上_其他_其他_0"),
                    Signal("15分钟_倒0笔_长度_任意_其他_其他_0"),
                ],
            )
        ],
    )
    m, f = event.is_match(s)
    assert m and f

    event = Event(
        name="单测",
        operate=Operate.LO,
        factors=[
            Factor(
                name="测试",
                signals_all=[
                    Signal("15分钟_倒0笔_方向_向上_其他_其他_20"),
                    Signal("15分钟_倒0笔_长度_任意_其他_其他_0"),
                ],
            )
        ],
    )
    m, f = event.is_match(s)
    assert not m and not f

    event = Event(
        name="单测",
        operate=Operate.LO,
        factors=[
            Factor(
                name="测试",
                signals_all=[
                    Signal("15分钟_倒0笔_方向_向下_其他_其他_0"),
                    Signal("15分钟_倒0笔_长度_任意_其他_其他_0"),
                ],
            )
        ],
    )
    m, f = event.is_match(s)
    assert not m and not f

    event = Event.load(
        {
            "name": "开多",
            "operate": "开多",
            "signals_all": ["1分钟_D1_涨跌停V230331_任意_任意_任意_0", "1分钟_D0停顿分型_BE辅助V230106_看空_强_任意_0"],
            "signals_any": [],
            "signals_not": [],
            "factors": [
                {
                    "name": "SMA#40多头",
                    "signals_all": ["5分钟_D1#SMA#40MO10_BS辅助V230313_看多_任意_任意_0"],
                    "signals_any": [],
                    "signals_not": [],
                }
            ],
        }
    )
    assert len(event.get_signals_config()) == 3



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/2/24 16:17
describe: 数据工具
"""
from deprecated import deprecated
from .ts_cache import TsDataCache
from . import ts
from .base import *


@deprecated(version="1.0.0", reason="不再推荐使用，后续随整个 data 模块一起删除")
def get_symbols(dc: TsDataCache, step):
    """获取择时策略投研不同阶段对应的标的列表

    :param dc: 数据缓存
    :param step: 投研阶段
    :return:
    """
    stocks = dc.stock_basic()
    stocks_ = stocks[stocks["list_date"] < "2010-01-01"].ts_code.to_list()
    stocks_map = {
        "index": [
            "000905.SH",
            "000016.SH",
            "000300.SH",
            "000001.SH",
            "000852.SH",
            "399001.SZ",
            "399006.SZ",
            "399376.SZ",
            "399377.SZ",
            "399317.SZ",
            "399303.SZ",
        ],
        "stock": stocks.ts_code.to_list(),
        "check": ["000001.SZ"],
        "train": stocks_[:200],
        "valid": stocks_[200:600],
        "etfs": [
            "512880.SH",
            "518880.SH",
            "515880.SH",
            "513050.SH",
            "512690.SH",
            "512660.SH",
            "512400.SH",
            "512010.SH",
            "512000.SH",
            "510900.SH",
            "510300.SH",
            "510500.SH",
            "510050.SH",
            "159992.SZ",
            "159985.SZ",
            "159981.SZ",
            "159949.SZ",
            "159915.SZ",
        ],
    }

    asset_map = {"index": "I", "stock": "E", "check": "E", "train": "E", "valid": "E", "etfs": "FD"}
    asset = asset_map[step]
    symbols = [f"{ts_code}#{asset}" for ts_code in stocks_map[step]]
    return symbols



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/2/26 15:03
describe: 使用 Plotly 构建绘图模块
"""
import os
import numpy as np
import pandas as pd
from plotly import graph_objects as go


class KlineChart:
    """K线绘图工具类

    plotly 参数详解: https://www.jianshu.com/p/4f4daf47cc85

    """

    def __init__(self, n_rows=3, **kwargs):
        """K线绘图工具类

        初始化执行逻辑：

        - 接收一个可选参数 n_rows，默认值为 3。这个参数表示图表中的子图数量。
        - 接收一个可变参数列表 **kwargs，可以传递其他配置参数。
        - 如果没有提供 row_heights 参数，则根据 n_rows 设置默认的行高度。
        - 定义了一些颜色变量：color_red 和 color_green。
        - 使用 make_subplots 函数创建一个具有 n_rows 行和 1 列的子图布局，并设置一些共享属性和间距。
        - 使用 fig.update_yaxes 和 fig.update_xaxes 更新 Y 轴和 X 轴的属性，如显示网格、自动调整范围等。
        - 使用 fig.update_layout 更新整个图形的布局，包括标题、边距、图例位置和样式、背景模板等。
        - 将 fig 对象保存在 self.fig 属性中。

        :param n_rows: 子图数量
        :param kwargs:
        """
        from plotly.subplots import make_subplots

        self.n_rows = n_rows
        row_heights = kwargs.get("row_heights", None)
        if not row_heights:
            heights_map = {3: [0.6, 0.2, 0.2], 4: [0.55, 0.15, 0.15, 0.15], 5: [0.4, 0.15, 0.15, 0.15, 0.15]}
            assert self.n_rows in heights_map.keys(), "使用内置高度配置，n_rows 只能是 3, 4, 5"
            row_heights = heights_map[self.n_rows]

        self.color_red = "rgba(249,41,62,0.7)"
        self.color_green = "rgba(0,170,59,0.7)"
        fig = make_subplots(
            rows=self.n_rows,
            cols=1,
            shared_xaxes=True,
            row_heights=row_heights,
            horizontal_spacing=0,
            vertical_spacing=0,
        )

        fig = fig.update_yaxes(
            showgrid=True,
            zeroline=False,
            automargin=True,
            fixedrange=kwargs.get("y_fixed_range", True),
            showspikes=True,
            spikemode="across",
            spikesnap="cursor",
            showline=False,
            spikedash="dot",
        )
        fig = fig.update_xaxes(
            type="category",
            rangeslider_visible=False,
            showgrid=False,
            automargin=True,
            showticklabels=False,
            showspikes=True,
            spikemode="across",
            spikesnap="cursor",
            showline=False,
            spikedash="dot",
        )

        # https://plotly.com/python/reference/layout/
        fig.update_layout(
            title=dict(text=kwargs.get("title", ""), yanchor="top"),
            margin=go.layout.Margin(l=0, r=0, b=0, t=0),  # left margin  # right margin  # bottom margin  # top margin
            # https://plotly.com/python/reference/layout/#layout-legend
            legend=dict(orientation="h", yanchor="top", y=1.05, xanchor="left", x=0, bgcolor="rgba(0,0,0,0)"),
            template="plotly_dark",
            hovermode="x unified",
            hoverlabel=dict(bgcolor="rgba(255,255,255,0.1)", font=dict(size=20)),  # 透明，更容易看清后面k线
            dragmode="pan",
            legend_title_font_color="red",
            height=kwargs.get("height", 300),
        )

        self.fig = fig

    def add_kline(self, kline: pd.DataFrame, name: str = "K线", **kwargs):
        """绘制K线

        函数执行逻辑：

        1. 检查 kline 数据框是否包含 'text' 列。如果没有，则添加一个空字符串列。
        2. 使用 go.Candlestick 创建一个K线图，并传入以下参数：
            - x: 日期时间数据
            - open, high, low, close: 开盘价、最高价、最低价和收盘价
            - text: 显示在每个 K 线上的文本标签
            - name: 图例名称
            - showlegend: 是否显示图例
            - increasing_line_color 和 decreasing_line_color: 上涨时的颜色和下跌时的颜色
            - increasing_fillcolor 和 decreasing_fillcolor: 上涨时填充颜色和下跌时填充颜色
            - **kwargs: 可以传递其他自定义参数给 Candlestick 函数。

        3. 将创建的烛台图对象添加到 self.fig 中的第一个子图（row=1, col=1）。
        4. 使用 fig.update_traces 更新所有 traces 的 xaxis 属性为 "x1"。
        """
        if "text" not in kline.columns:
            kline["text"] = ""

        candle = go.Candlestick(
            x=kline["dt"],
            open=kline["open"],
            high=kline["high"],
            low=kline["low"],
            close=kline["close"],
            text=kline["text"],
            name=name,
            showlegend=True,
            increasing_line_color=self.color_red,
            decreasing_line_color=self.color_green,
            increasing_fillcolor=self.color_red,
            decreasing_fillcolor=self.color_green,
            **kwargs,
        )
        self.fig.add_trace(candle, row=1, col=1)
        self.fig.update_traces(xaxis="x1")

    def add_vol(self, kline: pd.DataFrame, row=2, **kwargs):
        """绘制成交量图

        函数执行逻辑：

        1. 首先，复制输入的 kline 数据框到 df。
        2. 使用 np.where 函数根据收盘价（df['close']）和开盘价（df['open']）之间的关系为 df 创建一个新列 'vol_color'。
           如果收盘价大于开盘价，则使用红色（self.color_red），否则使用绿色（self.color_green）。
        3. 调用 add_bar_indicator 方法绘制成交量图。传递以下参数：
            - x: 日期时间数据
            - y: 成交量数据
            - color: 根据 'vol_color' 列的颜色
            - name: 图例名称
            - row: 指定要添加指标的子图行数，默认值为 2
            - show_legend: 是否显示图例，默认值为 False
        """
        df = kline.copy()
        df["vol_color"] = np.where(df["close"] > df["open"], self.color_red, self.color_green)
        self.add_bar_indicator(df["dt"], df["vol"], color=df["vol_color"], name="成交量", row=row, show_legend=False)

    def add_sma(self, kline: pd.DataFrame, row=1, ma_seq=(5, 10, 20), visible=False, **kwargs):
        """绘制均线图

        函数执行逻辑：

        1. 复制输入的 kline 数据框到 df。
        2. 获取自定义参数 line_width，默认值为 0.6。
        3. 遍历 ma_seq 中的所有均线周期：
            - 对每个周期使用 pandas rolling 方法计算收盘价的移动平均线。
            - 调用 add_scatter_indicator 方法将移动平均线数据绘制为折线图。传递以下参数：
                - x: 日期时间数据
                - y: 移动平均线数据
                - name: 图例名称，格式为 "MA{ma}"，其中 {ma} 是当前的均线周期。
                - row: 指定要添加指标的子图行数，默认值为 1
                - line_width: 线宽，默认值为 0.6
                - visible: 是否可见，默认值为 False
                - show_legend: 是否显示图例，默认值为 True
        """
        df = kline.copy()
        line_width = kwargs.get("line_width", 0.6)
        for ma in ma_seq:
            self.add_scatter_indicator(
                df["dt"],
                df["close"].rolling(ma).mean(),
                name=f"MA{ma}",
                row=row,
                line_width=line_width,
                visible=visible,
                show_legend=True,
            )

    def add_macd(self, kline: pd.DataFrame, row=3, **kwargs):
        """绘制MACD图

        函数执行逻辑：

        1. 首先，复制输入的 kline 数据框到 df。
        2. 获取自定义参数 fastperiod、slowperiod 和 signalperiod。这些参数分别对应于计算 MACD 时使用的快周期、慢周期和信号周期，默认值分别为 12、26 和 9。
        3. 使用 talib 库的 MACD 函数计算 MACD 值（diff, dea, macd）。
        4. 创建一个名为 macd_colors 的 numpy 数组，根据 macd 值大于零的情况设置颜色：大于零使用红色（self.color_red），否则使用绿色（self.color_green）。
        5. 调用 add_scatter_indicator 方法将 diff 和 dea 绘制为折线图。传递以下参数：
            - x: 日期时间数据
            - y: diff 或 dea 数据
            - name: 图例名称，分别为 "DIFF" 和 "DEA"
            - row: 指定要添加指标的子图行数，默认值为 3
            - line_color: 线的颜色，分别为 'white' 和 'yellow'
            - show_legend: 是否显示图例，默认值为 False
            - line_width: 线宽，默认值为 0.6
        6. 调用 add_bar_indicator 方法将 macd 绘制为柱状图。传递以下参数：
            - x: 日期时间数据
            - y: macd 数据
            - name: 图例名称，为 "MACD"
            - row: 指定要添加指标的子图行数，默认值为 3
            - color: 根据 macd_colors 设置颜色
            - show_legend: 是否显示图例，默认值为 False
        """
        df = kline.copy()
        fastperiod = kwargs.get("fastperiod", 12)
        slowperiod = kwargs.get("slowperiod", 26)
        signalperiod = kwargs.get("signalperiod", 9)
        line_width = kwargs.get("line_width", 0.6)

        if "DIFF" in df.columns and "DEA" in df.columns and "MACD" in df.columns:
            diff, dea, macd = df["DIFF"], df["DEA"], df["MACD"]
        else:
            from czsc.utils.ta import MACD

            diff, dea, macd = MACD(df["close"], fastperiod=fastperiod, slowperiod=slowperiod, signalperiod=signalperiod)

        macd_colors = np.where(macd > 0, self.color_red, self.color_green)
        self.add_scatter_indicator(
            df["dt"], diff, name="DIFF", row=row, line_color="white", show_legend=False, line_width=line_width
        )
        self.add_scatter_indicator(
            df["dt"], dea, name="DEA", row=row, line_color="yellow", show_legend=False, line_width=line_width
        )
        self.add_bar_indicator(df["dt"], macd, name="MACD", row=row, color=macd_colors, show_legend=False)

    def add_indicator(
        self, dt, scatters: list = None, scatter_names: list = None, bar=None, bar_name="", row=4, **kwargs
    ):
        """绘制曲线叠加bar型指标

        1. 获取自定义参数 line_width，默认值为 0.6。
        2. 如果 scatters（列表）不为空，则遍历 scatters 中的所有散点数据：
            - 对于每个散点数据，调用 add_scatter_indicator 方法将其绘制为折线图。传递以下参数：
                - x: 日期时间数据
                - y: 散点数据
                - name: 图例名称，来自 scatter_names 列表
                - row: 指定要添加指标的子图行数，默认值为 4
                - show_legend: 是否显示图例，默认值为 False
                - line_width: 线宽，默认值为 0.6
        3. 如果 bar 不为空，则使用 np.where 函数根据 bar 值大于零的情况设置颜色：大于零使用红色（self.color_red），否则使用绿色（self.color_green）。
        4. 调用 add_bar_indicator 方法将 bar 绘制为柱状图。传递以下参数：
            - x: 日期时间数据
            - y: bar 数据
            - name: 图例名称，为传入的 bar_name 参数
            - row: 指定要添加指标的子图行数，默认值为 4
            - color: 根据上一步计算的颜色设置
            - show_legend: 是否显示图例，默认值为 False
        """
        line_width = kwargs.get("line_width", 0.6)
        for i, scatter in enumerate(scatters):
            self.add_scatter_indicator(
                dt, scatter, name=scatter_names[i], row=row, show_legend=False, line_width=line_width
            )

        if bar:
            bar_colors = np.where(np.array(bar, dtype=np.double) > 0, self.color_red, self.color_green)
            self.add_bar_indicator(dt, bar, name=bar_name, row=row, color=bar_colors, show_legend=False)

    def add_marker_indicator(self, x, y, name: str, row: int, text=None, **kwargs):
        """绘制标记类指标

        函数执行逻辑：

        1. 获取自定义参数 line_color、line_width、hover_template、show_legend 和 visible。
            这些参数分别对应于折线颜色、宽度、鼠标悬停时显示的模板、是否显示图例和是否可见。
        2. 使用给定的 x、y 数据创建一个 go.Scatter 对象（散点图），并传入以下参数：
            - x: 指标的x轴数据
            - y: 指标的y轴数据
            - name: 指标名称
            - text: 文本说明
            - line_width: 线宽
            - line_color: 线颜色
            - hovertemplate: 鼠标悬停时显示的模板
            - showlegend: 是否显示图例
            - visible: 是否可见
            - opacity: 透明度
            - mode: 绘制模式，为 'markers' 表示只绘制标记
            - marker: 标记的样式，包括大小、颜色和符号
        3. 调用 self.fig.add_trace 方法将创建的 go.Scatter 对象添加到指定子图中，并更新所有 traces 的 X 轴属性为 "x1"。

        :param x: 指标的x轴
        :param y: 指标的y轴
        :param name: 指标名称
        :param row: 放入第几个子图
        :param text: 文本说明
        :param kwargs:
        :return:
        """
        line_color = kwargs.get("line_color", None)
        line_width = kwargs.get("line_width", None)
        hover_template = kwargs.get("hover_template", "%{y:.3f}-%{text}")
        show_legend = kwargs.get("show_legend", True)
        visible = True if kwargs.get("visible", True) else "legendonly"
        color = kwargs.get("color", None)
        tag = kwargs.get("tag", None)
        scatter = go.Scatter(
            x=x,
            y=y,
            name=name,
            text=text,
            line_width=line_width,
            line_color=line_color,
            hovertemplate=hover_template,
            showlegend=show_legend,
            visible=visible,
            opacity=1.0,
            mode="markers",
            marker=dict(size=10, color=color, symbol=tag),
        )

        self.fig.add_trace(scatter, row=row, col=1)
        self.fig.update_traces(xaxis="x1")

    def add_scatter_indicator(self, x, y, name: str, row: int, text=None, **kwargs):
        """绘制线性/离散指标

        绘图API文档：https://plotly.com/python-api-reference/generated/plotly.graph_objects.Scatter.html

        函数执行逻辑：

        1. 获取自定义参数 mode、hover_template、show_legend、opacity 和 visible。这些参数分别对应于绘图模式、鼠标悬停时显示的模板、是否显示图例、透明度和是否可见。
        2. 使用给定的 x、y 数据创建一个 go.Scatter 对象（散点图），并传入以下参数：
            - x: 指标的x轴数据
            - y: 指标的y轴数据
            - name: 指标名称
            - text: 文本说明
            - mode: 绘制模式，默认为 'text+lines'，表示同时绘制文本和线条
            - hovertemplate: 鼠标悬停时显示的模板
            - showlegend: 是否显示图例
            - visible: 是否可见
            - opacity: 透明度
        3. 调用 self.fig.add_trace 方法将创建的 go.Scatter 对象添加到指定子图中，并更新所有 traces 的 X 轴属性为 "x1"。

        :param x: 指标的x轴
        :param y: 指标的y轴
        :param name: 指标名称
        :param row: 放入第几个子图
        :param text: 文本说明
        :param kwargs:
        :return:
        """
        mode = kwargs.pop("mode", "text+lines")
        hover_template = kwargs.pop("hover_template", "%{y:.3f}")
        show_legend = kwargs.pop("show_legend", True)
        opacity = kwargs.pop("opacity", 1.0)
        visible = True if kwargs.pop("visible", True) else "legendonly"

        scatter = go.Scatter(
            x=x,
            y=y,
            name=name,
            text=text,
            mode=mode,
            hovertemplate=hover_template,
            showlegend=show_legend,
            visible=visible,
            opacity=opacity,
            **kwargs,
        )
        self.fig.add_trace(scatter, row=row, col=1)
        self.fig.update_traces(xaxis="x1")

    def add_bar_indicator(self, x, y, name: str, row: int, color=None, **kwargs):
        """绘制条形图指标

        绘图API文档：https://plotly.com/python-api-reference/generated/plotly.graph_objects.Bar.html

        函数执行逻辑：

        1. 获取自定义参数 hover_template、show_legend、visible 和 base。这些参数分别对应于鼠标悬停时显示的模板、是否显示图例、是否可见和基线（默认为 True）。
        2. 如果 color 参数为空，则使用 self.color_red 作为颜色。
        3. 使用给定的 x、y 数据创建一个 go.Bar 对象（条形图），并传入以下参数：
            - x: 指标的x轴数据
            - y: 指标的y轴数据
            - marker_line_color: 条形边框的颜色
            - marker_color: 条形填充的颜色
            - name: 指标名称
            - showlegend: 是否显示图例
            - hovertemplate: 鼠标悬停时显示的模板
            - visible: 是否可见
            - base: 基线，默认为 True
        4. 调用 self.fig.add_trace 方法将创建的 go.Bar 对象添加到指定子图中，并更新所有 traces 的 X 轴属性为 "x1"。

        :param x: 指标的x轴
        :param y: 指标的y轴
        :param name: 指标名称
        :param row: 放入第几个子图
        :param color: 指标的颜色，可以是单个颜色，也可以是一个列表，列表长度和y的长度一致，指示每个y的颜色
            比如：color = 'rgba(249,41,62,0.7)' 或者 color = ['rgba(249,41,62,0.7)', 'rgba(0,170,59,0.7)']
        :param kwargs:
        :return:
        """
        hover_template = kwargs.pop("hover_template", "%{y:.3f}")
        show_legend = kwargs.pop("show_legend", True)
        visible = kwargs.pop("visible", True)
        base = kwargs.pop("base", True)
        if color is None:
            color = self.color_red

        bar = go.Bar(
            x=x,
            y=y,
            marker_line_color=color,
            marker_color=color,
            name=name,
            showlegend=show_legend,
            hovertemplate=hover_template,
            visible=visible,
            base=base,
            **kwargs,
        )
        self.fig.add_trace(bar, row=row, col=1)
        self.fig.update_traces(xaxis="x1")

    def open_in_browser(self, file_name: str = None, **kwargs):
        """在浏览器中打开"""
        import webbrowser

        if not file_name:
            from czsc.utils.cache import home_path

            file_name = os.path.join(home_path, "kline_chart.html")

        self.fig.update_layout(**kwargs)
        self.fig.write_html(file_name)
        webbrowser.open(file_name)


def plot_nx_graph(g, **kwargs) -> go.Figure:
    """使用 Plotly 绘制 nx.Graph 的图形

    :param g: nx.Graph 对象
    :param kwargs:
    :return: go.Figure 对象
    """
    import networkx as nx

    title = kwargs.get("title", "Network graph made with Python")
    edge_width = kwargs.get("edge_width", 1.5)
    node_marker_size = kwargs.get("node_marker_size", 10)

    # 使用 spring_layout 为图分配位置
    pos = nx.spring_layout(g)

    # 准备绘图数据
    edge_x = []
    edge_y = []
    edge_weights = []
    for edge in g.edges():
        x0, y0 = pos[edge[0]]
        x1, y1 = pos[edge[1]]
        edge_x.extend([x0, x1, None])
        edge_y.extend([y0, y1, None])
        edge_weights.append(f'{g[edge[0]][edge[1]]["weight"]:.2f}')

    node_x = []
    node_y = []
    node_labels = []
    for node in g.nodes():
        node_x.append(pos[node][0])
        node_y.append(pos[node][1])
        node_labels.append(node)

    # 创建边的散点图
    edge_trace = go.Scatter(
        x=edge_x,
        y=edge_y,
        line=dict(width=edge_width, color="#888"),
        hoverinfo="none",
        mode="lines",
    )

    # 创建节点的散点图
    node_trace = go.Scatter(
        x=node_x,
        y=node_y,
        mode="markers",
        hoverinfo="text",
        text=node_labels,  # 添加节点标签
        marker=dict(
            showscale=False,
            color="skyblue",
            size=node_marker_size,
            line_width=0,
        ),
    )

    # 计算边的中点位置并添加注释
    edge_annotations = []
    for edge in g.edges():
        x0, y0 = pos[edge[0]]
        x1, y1 = pos[edge[1]]
        mid_x = (x0 + x1) / 2
        mid_y = (y0 + y1) / 2
        weight = f'{g[edge[0]][edge[1]]["weight"]:.2f}'
        edge_annotations.append(
            dict(
                x=mid_x,
                y=mid_y,
                text=weight,
                showarrow=False,
                font=dict(size=12, color="red" if float(weight) > 0 else "green"),
            )
        )

    # 创建图表
    fig = go.Figure(
        data=[edge_trace, node_trace],
        layout=go.Layout(
            title=f"<br>{title}",
            titlefont_size=16,
            showlegend=False,
            hovermode="closest",
            margin=dict(b=20, l=5, r=5, t=40),
            xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
            yaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
            annotations=edge_annotations,  # 添加边的注释
        ),
    )

    return fig



import pandas as pd
import numpy as np


def test_daily_performance():
    from czsc import daily_performance

    x = daily_performance([0.01, 0.02, -0.01, 0.03, 0.02, -0.02, 0.01, -0.01, 0.02, 0.01])
    print(x)


def test_weight_backtest():
    from czsc import WeightBacktest
    from test.test_analyze import read_daily

    dfw = read_daily()
    dfw = pd.DataFrame(dfw)
    dfw["weight"] = np.where(dfw["close"] > dfw["open"], 1, -1)
    dfw["price"] = dfw["close"]
    wb = WeightBacktest(dfw[["dt", "weight", "symbol", "price"]])

    assert wb.stats["夏普"] == -0.0433



import pytest
import numpy as np
import pandas as pd


def test_is_event_feature():
    from czsc.features.utils import is_event_feature

    # 测试事件类因子
    df1 = pd.DataFrame({'factor': [0, 1, -1, 0, 1, -1]})
    assert is_event_feature(df1, 'factor') is True

    # 测试非事件类因子
    df2 = pd.DataFrame({'factor': [0, 1, 2, 3, 4, 5]})
    assert is_event_feature(df2, 'factor') is False


def test_rolling_tanh():
    from czsc.features.utils import rolling_tanh

    # Create a dummy dataframe
    df = pd.DataFrame({
        'dt': pd.date_range(start='1/1/2021', periods=500),
        'col1': np.random.rand(500)
    })

    # Apply the rolling_tanh function
    result_df = rolling_tanh(df, 'col1')
    assert 'col1_tanh' in result_df.columns
    assert result_df['col1_tanh'].between(-1, 1).all()

    # Apply the rolling_tanh function
    result_df = rolling_tanh(df, 'col1', new_col='col1_tanh2')
    assert 'col1_tanh2' in result_df.columns
    assert result_df['col1_tanh2'].between(-1, 1).all()

    result_df = rolling_tanh(df, 'col1', new_col='col1_tanh3', window=100, min_periods=50)
    assert 'col1_tanh3' in result_df.columns
    assert result_df['col1_tanh3'].between(-1, 1).all()


def test_normalize_corr():
    from czsc.features.utils import normalize_corr

    np.random.seed(123)
    # Create a fake DataFrame
    df = pd.DataFrame({
        'dt': pd.date_range(start='1/1/2021', periods=3000),
        'symbol': ['AAPL'] * 3000,
        'price': np.random.rand(3000),
        'factor': np.random.rand(3000),
    })

    df['n1b'] = df['price'].shift(-1) / df['price'] - 1
    raw_corr = df['n1b'].corr(df['factor'])

    # Call the function with the fake DataFrame
    result = normalize_corr(df, fcol='factor', copy=True, mode='rolling', window=600)
    corr1 = result['n1b'].corr(result['factor'])
    assert result.shape == df.shape and np.sign(corr1) == -np.sign(raw_corr)

    # Call the function with the fake DataFrame
    result = normalize_corr(df, fcol='factor', copy=True, mode='rolling', window=300)
    corr1 = result['n1b'].corr(result['factor'])
    assert result.shape == df.shape and np.sign(corr1) == np.sign(raw_corr)

    result = normalize_corr(df, fcol='factor', copy=True, mode='rolling', window=2000)
    corr1 = result['n1b'].corr(result['factor'])
    assert result.shape == df.shape and np.sign(corr1) == -np.sign(raw_corr)

    result = normalize_corr(df, fcol='factor', copy=True, mode='simple')
    corr2 = result['n1b'].corr(result['factor'])
    assert result.shape == df.shape and corr2 == -raw_corr



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/11/17 18:50
"""
import pandas as pd
import numpy as np
from tqdm import tqdm
from collections import Counter
from typing import List
from deprecated import deprecated
from sklearn.preprocessing import KBinsDiscretizer
from ..data import TsDataCache


def discretizer(df: pd.DataFrame, col: str, n_bins=20, encode="ordinal", strategy="quantile"):
    """使用 KBinsDiscretizer 对连续变量在时间截面上进行离散化

    :param df: 数据对象
    :param col: 连续变量列名
    :param n_bins: 参见 KBinsDiscretizer 文档
        https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.KBinsDiscretizer.html
    :param encode: 参见 KBinsDiscretizer 文档
    :param strategy: 参见 KBinsDiscretizer 文档
    :return:
    """
    assert col in df.columns, f"{col} not in {df.columns}"
    assert "dt" in df.columns

    new_col = f"{col}_bins{n_bins}"
    results = []
    for dt, dfg in tqdm(df.groupby("dt"), desc=f"{col}_bins{n_bins}"):
        kb = KBinsDiscretizer(n_bins=n_bins, encode=encode, strategy=strategy)
        # 加1，使分组从1开始
        dfg[new_col] = kb.fit_transform(dfg[col].values.reshape(-1, 1)).ravel() + 1
        results.append(dfg)
    df = pd.concat(results, ignore_index=True)
    return df


def get_index_beta(dc: TsDataCache, sdt: str, edt: str, freq="D", file_xlsx=None, indices=None):
    """获取基准指数的Beta

    :param dc: 数据缓存对象
    :param sdt: 开始日期
    :param edt: 结束日期
    :param freq: K线周期，D 日线，W 周线，M 月线
    :param file_xlsx: 结果保存文件
    :param indices: 定义指数列表
    :return:
    """
    if not indices:
        indices = ["000001.SH", "000016.SH", "000905.SH", "000300.SH", "399001.SZ", "399006.SZ"]

    beta = {}
    p = []
    for ts_code in indices:
        df = dc.pro_bar(ts_code=ts_code, start_date=sdt, end_date=edt, freq=freq, asset="I", raw_bar=False)
        beta[ts_code] = df
        df = df.fillna(0)
        start_i, end_i, mdd = max_draw_down(df["n1b"].to_list())
        start_dt = df.iloc[start_i]["trade_date"]
        end_dt = df.iloc[end_i]["trade_date"]
        row = {
            "标的": ts_code,
            "开始日期": sdt,
            "结束日期": edt,
            "最大回撤": mdd,
            "回撤开始": start_dt,
            "回撤结束": end_dt,
            "交易次数": len(df),
            "交易胜率": round(len(df[df.n1b > 0]) / len(df), 4),
            "累计收益": round(df.n1b.sum(), 4),
        }
        cols = [x for x in df.columns if x[0] == "n" and x[-1] == "b"]
        row.update({x: round(df[x].mean(), 4) for x in cols})
        p.append(row)

    dfp = pd.DataFrame(p)
    if file_xlsx:
        f = pd.ExcelWriter(file_xlsx)
        dfp.to_excel(f, index=False, sheet_name="指数表现")
        for name, df_ in beta.items():
            df_.to_excel(f, index=False, sheet_name=name)
        f.close()
    else:
        beta["dfp"] = dfp
        return beta


def max_draw_down(n1b: List):
    """最大回撤

    参考：https://blog.csdn.net/weixin_38997425/article/details/82915386

    :param n1b: 逐个结算周期的收益列表，单位：BP，换算关系是 10000BP = 100%
        如，n1b = [100.1, -90.5, 212.6]，表示第一个结算周期收益为100.1BP，也就是1.001%，以此类推。
    :return: 最大回撤起止位置和最大回撤
    """
    curve = np.cumsum(n1b)
    curve += 10000
    # 获取结束位置
    i = np.argmax((np.maximum.accumulate(curve) - curve) / np.maximum.accumulate(curve))
    if i == 0:
        return 0, 0, 0

    # 获取开始位置
    j = np.argmax(curve[:i])
    mdd = int((curve[j] - curve[i]) / curve[j] * 10000) / 10000
    return j, i, mdd


def turn_over_rate(df_holds: pd.DataFrame) -> [pd.DataFrame, float]:
    """计算持仓明细对应的组合换手率

    :param df_holds: 每个交易日的持仓明细，数据样例如下
                证券代码    成分日期    持仓权重
            0  000576.SZ  2020-01-02  0.0099
            1  000639.SZ  2020-01-02  0.0099
            2  000803.SZ  2020-01-02  0.0099
            3  000811.SZ  2020-01-02  0.0099
            4  000829.SZ  2020-01-02  0.0099
    :return: 组合换手率
    """
    dft = pd.pivot_table(df_holds, index="成分日期", columns="证券代码", values="持仓权重", aggfunc="sum")
    dft = dft.fillna(0)
    df_turns = dft.diff().abs().sum(axis=1).reset_index()
    df_turns.columns = ["date", "change"]

    # 由于是 diff 计算，第一个时刻的仓位变化被忽视了，修改一下
    sdt = df_holds["成分日期"].min()
    df_turns.loc[(df_turns["date"] == sdt), "change"] = df_holds[df_holds["成分日期"] == sdt]["持仓权重"].sum()
    return df_turns, round(df_turns.change.sum() / 2, 4)


def holds_concepts_effect(holds: pd.DataFrame, concepts: dict, top_n=20, min_n=3, **kwargs):
    """股票持仓列表的板块效应

    原理概述：在选股时，如果股票的概念板块与组合中的其他股票的概念板块有重合，那么这个股票的表现会更好。

    函数计算逻辑:

    1. 如果kwargs中存在'copy'键且对应值为True，则将holds进行复制。
    2. 为holds添加'概念板块'列，该列的值是holds中'symbol'列对应的股票的概念板块列表，如果没有对应的概念板块则填充为空。
    3. 添加'概念数量'列，该列的值是每个股票的概念板块数量。
    4. 从holds中筛选出概念数量大于0的行，赋值给holds。
    5. 创建空列表new_holds和空字典dt_key_concepts。
    6. 对holds按照'dt'进行分组，遍历每个分组，计算板块效应。
    a. 计算密集出现的概念，选取出现次数最多的前top_n个概念，赋值给key_concepts列表。
    b. 将日期dt和对应的key_concepts存入dt_key_concepts字典。
    c. 计算在密集概念中出现次数超过min_n的股票，将符合条件的股票添加到new_holds列表中。
    7. 使用pd.concat将new_holds中的DataFrame进行合并，忽略索引，赋值给dfh。
    8. 创建DataFrame dfk，其中包含日期(dt)和对应的强势概念(key_concepts)。
    9. 返回dfh和dfk。

    :param holds: 组合股票池数据，样例：

            ===================  =========  ==========
            dt                   symbol         weight
            ===================  =========  ==========
            2023-05-09 00:00:00  601858.SH  0.00333333
            2023-05-09 00:00:00  300502.SZ  0.00333333
            2023-05-09 00:00:00  603258.SH  0.00333333
            2023-05-09 00:00:00  300499.SZ  0.00333333
            2023-05-09 00:00:00  300624.SZ  0.00333333
            ===================  =========  ==========

    :param concepts: 股票的概念板块，样例：
            {
                '002507.SZ': ['电子商务', '超级品牌', '国企改革'],
                '002508.SZ': ['家用电器', '杭州亚运会', '恒大概念']
            }
    :param top_n: 选取前 n 个密集概念
    :param min_n: 单股票至少要有 n 个概念在 top_n 中
    :return: 过滤后的选股结果，每个时间点的 top_n 概念
    """
    if kwargs.get("copy", True):
        holds = holds.copy()

    holds["概念板块"] = holds["symbol"].map(concepts).fillna("")
    holds["概念数量"] = holds["概念板块"].apply(len)
    holds = holds[holds["概念数量"] > 0]

    new_holds = []
    dt_key_concepts = {}
    for dt, dfg in tqdm(holds.groupby("dt"), desc="计算板块效应"):
        # 计算密集出现的概念
        key_concepts = [k for k, v in Counter([x for y in dfg["概念板块"] for x in y]).most_common(top_n)]
        dt_key_concepts[dt] = key_concepts

        # 计算在密集概念中出现次数超过min_n的股票
        dfg["强势概念"] = dfg["概念板块"].apply(lambda x: ",".join(set(x) & set(key_concepts)))
        sel = dfg[dfg["强势概念"].apply(lambda x: len(x.split(",")) >= min_n)]
        new_holds.append(sel)

    dfh = pd.concat(new_holds, ignore_index=True)
    dfk = pd.DataFrame([{"dt": k, "强势概念": v} for k, v in dt_key_concepts.items()])
    return dfh, dfk



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/7/7 22:42
describe: Event 相关的传感器
"""
import os
import shutil
import pandas as pd
from copy import deepcopy
from loguru import logger
from czsc.objects import Event
from typing import List, Dict, Callable, Any, Union
from czsc.traders.sig_parse import get_signals_freqs
from czsc.traders.base import generate_czsc_signals
from czsc.utils.io import save_json
from concurrent.futures import ProcessPoolExecutor, as_completed


class EventMatchSensor:
    def __init__(self, events: List[Union[Dict[str, Any], Event]], symbols: List[str], read_bars: Callable, **kwargs) -> None:
        """
        事件匹配传感器

        :param event: 事件配置
        :param symbols: 事件匹配的标的
        :param read_bars: 读取K线数据的函数，函数签名如下：
            read_bars(symbol, freq, sdt, edt, fq='前复权', **kwargs) -> List[RawBar]

        :param kwargs: 读取K线数据的函数的参数

            - bar_sdt: K线数据的开始时间，格式：2020-01-01
            - sdt: 事件匹配的开始时间，格式：2020-01-01
            - edt: 事件匹配的结束时间，格式：2020-01-01
            - max_workers: 读取K线数据的函数的最大进程数
            - signals_module: 信号解析模块，如：czsc.signals
            - results_path: 事件匹配结果的保存路径

        """
        self.symbols = symbols
        self.read_bars = read_bars
        self.events = [Event.load(event) if isinstance(event, dict) else event for event in events]
        self.events_map = {event.name: event for event in self.events}
        self.events_name = [event.name for event in self.events]
        self.results_path = kwargs.pop("results_path")
        self.refresh = kwargs.pop("refresh", False)
        if os.path.exists(self.results_path) and self.refresh:
            shutil.rmtree(self.results_path)
            logger.warning(f"文件夹 {self.results_path} 已存在，程序将覆盖该文件夹下的所有文件")
        os.makedirs(self.results_path, exist_ok=True)
        save_json({e.name: e.dump() for e in self.events}, os.path.join(self.results_path, "events.json"))

        logger.add(os.path.join(self.results_path, "event_match_sensor.log"), rotation="1 day", encoding="utf-8")
        logger.info(f"事件匹配传感器初始化，共有{len(self.events)}个事件，{len(self.symbols)}个标的")

        self.signals_module = kwargs.pop("signals_module", "czsc.signals")
        self.signals_config = self._get_signals_config()
        self.freqs = get_signals_freqs(self.signals_config)
        self.base_freq = self.freqs[0]
        logger.info(
            f"signals_moudle: {self.signals_module}, signals_config: {self.signals_config}, freqs: {self.freqs}"
        )

        self.bar_sdt = kwargs.pop("bar_sdt", "2017-01-01")
        self.sdt = kwargs.pop("sdt", "2018-01-01")
        self.edt = kwargs.pop("edt", "2022-01-01")
        logger.info(f"bar_sdt: {self.bar_sdt}, sdt: {self.sdt}, edt: {self.edt}")

        self.kwargs = kwargs
        logger.info(f"事件匹配传感器初始化完成，共有{len(self.events)}个事件，{len(self.symbols)}个标的")
        file_data = os.path.join(self.results_path, "data.feather")
        if os.path.exists(file_data):
            self.data = pd.read_feather(file_data)
            logger.info(f"读取事件匹配数据：{file_data}")
        else:
            self.data = self._multi_symbols(self.symbols, max_workers=self.kwargs.pop("max_workers", 1))
            self.data.to_feather(os.path.join(self.results_path, "data.feather"))

        _res = []
        for event_name in self.events_name:
            df = self.get_event_csc(event_name)
            df = df.set_index("dt")
            _res.append(df)
        df = pd.concat(_res, axis=1, ignore_index=False).reset_index()
        file_csc = os.path.join(self.results_path, "cross_section_counts.csv")
        df.to_csv(file_csc, index=False)
        logger.info(f"截面匹配次数计算完成，结果保存至：{file_csc}")

        # csc = cross section count，表示截面匹配次数
        self.csc = df

    def _get_signals_config(self):
        """获取所有事件的信号配置，并将其合并为一个不包含重复项的列表。
        该列表包含了所有事件所需的信号计算和解析规则，以便于后续的事件匹配过程。

        1. 创建一个空列表 config，用于存储所有的信号配置。
        2. 遍历 self.events 中的所有事件（Event 对象）。对于每个事件，调用其 get_signals_config 方法，
            传入 signals_module 参数，并将返回值（即该事件的信号配置）添加到 config 列表中。
        3. 通过 list comprehension 生成一个新的列表 config。对 config 列表中的每个字典 d
            使用 tuple(d.items()) 转换为元组，然后将这些元组转换回 dict 并加入新列表中。
        4. 返回处理后的 config 列表。
        """
        config = []
        for event in self.events:
            _c = event.get_signals_config(signals_module=self.signals_module)
            config.extend(_c)
        config = [dict(t) for t in {tuple(d.items()) for d in config}]
        return config

    def _single_symbol(self, symbol):
        """单个symbol的事件匹配

        对单个标的（symbol）进行事件匹配。它首先获取 K 线数据，然后生成 CZSC 信号，接着遍历每个事件并计算匹配情况，
        最后整理数据框并返回。如果在过程中遇到问题，则记录错误并返回一个空 DataFrame。

        函数执行逻辑：

        1. 调用 self.read_bars 方法读取指定 symbol、频率（self.base_freq）、开始时间（self.bar_sdt）
           和结束时间（self.edt）的 K 线数据，并将返回值赋给 bars。
        2. 使用 generate_czsc_signals 函数生成 CZSC 信号。这里传入了 bars、复制后的
           signals_config（以防止修改原始配置）、开始时间（self.sdt）以及 df=False（表示返回一个字典列表而非 DataFrame 对象）。
        3. 将上一步生成的信号转换为 DataFrame 并保存到 sigs 变量中。
        4. 创建一个新的 events 复制品（以防止修改原始事件列表），并创建一个空列表 new_cols，用于存储新添加的列名。
        5. 遍历新的 events 列表，对于每个 event：
            a. 获取 event 的名称 e_name。
            b. 使用 apply 函数应用 is_match 方法来判断每行数据是否与该事件相匹配。
                结果是一个布尔值和一个 float 值（表示匹配得分），它们分别被保存为 e_name 和 f'{e_name}_F' 列。
            c. 将这两个新列名添加到 new_cols 列表中。
        6. 在 sigs 数据框中添加一列 n1b，表示涨跌幅。
        7. 最后，重新组织 sigs 数据框的列顺序，使其包含以下列：symbol、dt、open、close、high、low、vol、amount、n1b 以及所有新添加的列。

        """
        try:
            bars = self.read_bars(symbol, freq=self.base_freq, sdt=self.bar_sdt, edt=self.edt, **self.kwargs)
            sigs = generate_czsc_signals(bars, deepcopy(self.signals_config), sdt=self.sdt, df=False)
            sigs = pd.DataFrame(sigs)
            events = deepcopy(self.events)
            new_cols = []
            for event in events:
                e_name = event.name
                sigs[[e_name, f'{e_name}_F']] = sigs.apply(event.is_match, axis=1, result_type="expand")  # type: ignore
                new_cols.extend([e_name, f'{e_name}_F'])
            sigs['n1b'] = (sigs['close'].shift(-1) / sigs['close'] - 1) * 10000
            sigs = sigs[['symbol', 'dt', 'open', 'close', 'high', 'low', 'vol', 'amount', 'n1b'] + new_cols]  # type: ignore
            return sigs
        except Exception as e:
            logger.error(f"{symbol} 事件匹配失败：{e}")
            return pd.DataFrame()

    def _multi_symbols(self, symbols: List[str], max_workers=1):
        """多个symbol的事件匹配"""
        logger.info(f"开始事件匹配，共有{len(symbols)}个标的，max_workers={max_workers}")
        dfs = []
        if max_workers == 1:
            for symbol in symbols:
                dfs.append(self._single_symbol(symbol))
        else:
            with ProcessPoolExecutor(max_workers) as executor:
                futures = [executor.submit(self._single_symbol, symbol) for symbol in symbols]
                for future in as_completed(futures):
                    dfs.append(future.result())
        df = pd.concat(dfs, ignore_index=True)
        for event_name in self.events_name:
            df[event_name] = df[event_name].astype(int)
        return df

    def get_event_csc(self, event_name: str):
        """获取事件的截面匹配次数

        csc = cross section count，表示截面匹配次数

        函数执行逻辑：

        1. 创建一个 self.data 的副本 df。
        2. 在 df 中筛选出 event_name 列等于 1 的行。
        3. 使用 groupby 方法按 symbol 和 dt 对筛选后的数据进行分组，并计算 event_name 列的总和。
            结果将形成一个新的 DataFrame，其中索引为 (symbol, dt) 组合，只有一个列 event_name，表示每个组合的匹配次数。
        4. 再次使用 groupby 方法按 dt 对上一步的结果进行分组，并计算 event_name 列的总和。这次得到的新 DataFrame
            只有一个列 event_name，表示在每个时间点所有标的的事件匹配总数。

        :param event_name: 事件名称
        :return: DataFrame
        """
        df = self.data.copy()
        df = df[df[event_name] == 1]
        df = df.groupby(["symbol", "dt"])[event_name].sum().reset_index()
        df = df.groupby("dt")[event_name].sum().reset_index()
        return df



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/11/14 12:39
describe: 从任意周期K线开始合成更高周期K线的工具类
"""
import pandas as pd
from datetime import datetime, timedelta, date
from typing import List, Union, AnyStr, Optional
from czsc.objects import RawBar, Freq
from pathlib import Path
from loguru import logger


mss = pd.read_feather(Path(__file__).parent / "minutes_split.feather")
freq_market_times, freq_edt_map = {}, {}
for _m, dfg in mss.groupby("market"):
    for _f in [x for x in mss.columns if x.endswith("分钟")]:
        freq_market_times[f"{_f}_{_m}"] = list(dfg[_f].unique())
        freq_edt_map[f"{_f}_{_m}"] = {k: v for k, v in dfg[["time", _f]].values}


def is_trading_time(dt: datetime = datetime.now(), market="A股"):
    """判断指定时间是否是交易时间"""
    hm = dt.strftime("%H:%M")
    times = freq_market_times[f"1分钟_{market}"]
    return True if hm in times else False


def get_intraday_times(freq="1分钟", market="A股"):
    """获取指定市场的交易时间段

    :param freq: K线周期，如 1分钟、5分钟、15分钟、30分钟、60分钟
    :param market: 市场名称，可选值：A股、期货、默认
    :return: 交易时间段列表
    """
    assert market in ["A股", "期货", "默认"], "market 参数必须为 A股 或 期货 或 默认"
    assert freq.endswith("分钟"), "freq 参数必须为分钟级别的K线周期"
    return freq_market_times[f"{freq}_{market}"]


def format_standard_kline(df: pd.DataFrame, freq: str):
    """格式化标准K线数据为 CZSC 标准数据结构 RawBar 列表

    :param df: 标准K线数据，DataFrame结构

        ===================  =========  ======  =======  ======  =====  ===========  ===========
        dt                   symbol       open    close    high    low          vol       amount
        ===================  =========  ======  =======  ======  =====  ===========  ===========
        2023-11-17 00:00:00  689009.SH   33.52    33.41   33.69  33.38  1.97575e+06  6.61661e+07
        2023-11-20 00:00:00  689009.SH   33.4     32.91   33.45  32.25  5.15016e+06  1.68867e+08
        ===================  =========  ======  =======  ======  =====  ===========  ===========

    :param freq: K线级别
    :return: list of RawBar
    """
    # from czsc.objects import RawBar, Freq
    bars = []
    for i, row in df.iterrows():
        bar = RawBar(
            id=i,
            symbol=row["symbol"],
            dt=row["dt"],
            open=row["open"],
            close=row["close"],
            high=row["high"],
            low=row["low"],
            vol=row["vol"],
            amount=row["amount"],
            freq=Freq(freq),
        )
        bars.append(bar)
    return bars


def check_freq_and_market(time_seq: List[AnyStr], freq: Optional[AnyStr] = None):
    """检查时间序列是否为同一周期，是否为同一市场

    函数计算逻辑：

    1. 如果`freq`在特定列表中，函数直接返回`freq`和"默认"作为市场类型。
    2. 如果`freq`是'1分钟'，函数会添加额外的时间点到`time_seq`中。
    3. 函数去除`time_seq`中的重复时间点，并确保其长度至少为2。
    4. 函数遍历`freq_market_times`字典，寻找与`time_seq`匹配的项，并返回对应的`freq_x`和`market`。
    5. 如果没有找到匹配的项，函数返回None和"默认"。

    :param time_seq: 时间序列，如 ['11:00', '15:00', '23:00', '01:00', '02:30']
    :param freq: 时间序列对应的K线周期，可选参数，使用该参数可以加快检查速度。
        可选值：1分钟、5分钟、15分钟、30分钟、60分钟、日线、周线、月线、季线、年线
    :return:
        - freq      K线周期
        - market    交易市场
    """
    if freq in ["日线", "周线", "月线", "季线", "年线"]:
        return freq, "默认"

    time_seq = sorted(list(set(time_seq)))
    assert len(time_seq) >= 2, "time_seq长度必须大于等于2"

    for key, tts in freq_market_times.items():
        if freq and not key.startswith(freq):
            continue
        freq_x, market = key.split("_")

        if freq_x == "1分钟":
            time_seq.extend(["14:57", "14:58", "14:59", "15:00"])

        sub_tts = [x for x in tts if x >= min(time_seq) and x <= max(time_seq)]
        if set(time_seq) == set(sub_tts):
            # print(f"check_freq_and_market: {freq_x} - {market}")
            return freq_x, market

    return None, "默认"


def freq_end_date(dt, freq: Union[Freq, AnyStr]):
    """交易日结束时间计算"""
    if not isinstance(dt, date):
        dt = pd.to_datetime(dt).date()
    if not isinstance(freq, Freq):
        freq = Freq(freq)

    dt = pd.to_datetime(dt)
    if freq == Freq.D:
        return dt

    if freq == Freq.W:
        return dt + timedelta(days=5 - dt.isoweekday())

    if freq == Freq.Y:
        edt = dt.replace(month=12, day=31)
        return edt

    if freq == Freq.M:
        if dt.month == 12:
            edt = dt.replace(year=dt.year + 1, month=1, day=1)
        else:
            edt = dt.replace(month=dt.month + 1, day=1)
        return edt - timedelta(days=1)

    if freq == Freq.S:
        dt_m = dt.month
        if dt_m in [1, 2, 3]:
            edt = dt.replace(month=4, day=1)
        elif dt_m in [4, 5, 6]:
            edt = dt.replace(month=7, day=1)
        elif dt_m in [7, 8, 9]:
            edt = dt.replace(month=10, day=1)
        else:
            edt = dt.replace(year=dt.year + 1, month=1, day=1)
        return edt - timedelta(days=1)

    logger.warning(f"error: {dt} - {freq}")
    return dt


def freq_end_time(dt: datetime, freq: Union[Freq, AnyStr], market="A股") -> datetime:
    """A股与期货市场精确的获取 dt 对应的K线周期结束时间

    :param dt: datetime
    :param freq: Freq
    :param market: str, A股 或 期货 或 默认
    :return: datetime
    """
    assert market in ["A股", "期货", "默认"], "market 参数必须为 A股 或 期货 或 默认"
    if not isinstance(freq, Freq):
        freq = Freq(freq)
    if dt.second > 0 or dt.microsecond > 0:
        dt = dt.replace(second=0, microsecond=0) + timedelta(minutes=1)

    hm = dt.strftime("%H:%M")
    key = f"{freq.value}_{market}"
    if freq.value.endswith("分钟"):
        h, m = map(int, freq_edt_map[key][hm].split(":"))
        edt = dt.replace(hour=h, minute=m)

        if h == m == 0 and freq != Freq.F1 and hm != "00:00":
            edt += timedelta(days=1)

        return edt

    # if not ("15:00" > hm > "09:00") and market == "期货":
    #     dt = next_trading_date(dt, n=1)

    return freq_end_date(dt.date(), freq)


def resample_bars(df: pd.DataFrame, target_freq: Union[Freq, AnyStr], raw_bars=True, **kwargs):
    """将给定的K线数据重新采样为目标周期的K线数据

    函数计算逻辑：

    1. 确定目标周期`target_freq`的类型和市场类型。
    2. 添加一个新列`freq_edt`，表示每个数据点对应的目标周期的结束时间。
    3. 根据`freq_edt`对数据进行分组，并对每组数据进行聚合，得到目标周期的K线数据。
    4. 重置索引，并选择需要的列。
    5. 根据`raw_bars`参数，决定返回的数据类型：如果为True，转换为`RawBar`对象；如果为False，直接返回DataFrame。
    6. 如果`drop_unfinished`参数为True，删除最后一根未完成的K线。

    :param df: 原始K线数据，必须包含以下列：symbol, dt, open, close, high, low, vol, amount。样例如下：
               symbol                  dt     open    close     high      low  \
        0  000001.XSHG 2015-01-05 09:31:00  3258.63  3259.69  3262.85  3258.63
        1  000001.XSHG 2015-01-05 09:32:00  3258.33  3256.19  3259.55  3256.19
        2  000001.XSHG 2015-01-05 09:33:00  3256.10  3257.50  3258.42  3256.10
        3  000001.XSHG 2015-01-05 09:34:00  3259.33  3261.76  3261.76  3257.98
        4  000001.XSHG 2015-01-05 09:35:00  3261.71  3264.88  3265.48  3261.71
                  vol        amount
        0  1333523100  4.346872e+12
        1   511386100  1.665170e+12
        2   455375200  1.483385e+12
        3   363393800  1.185303e+12
        4   402854600  1.315272e+12
    :param target_freq: 目标周期
    :param raw_bars: 是否将转换后的K线序列转换为RawBar对象
    :param kwargs:

        - base_freq: 基础周期，如果不指定，则根据df中的dt列自动推断
        - drop_unfinished: 是否删除最后一根未完成的K线

    :return: 转换后的K线序列
    """
    if not isinstance(target_freq, Freq):
        target_freq = Freq(target_freq)

    base_freq = kwargs.get("base_freq", None)
    if target_freq.value.endswith("分钟"):
        uni_times = sorted(df["dt"].tail(2000).apply(lambda x: x.strftime("%H:%M")).unique().tolist())
        _, market = check_freq_and_market(uni_times, freq=base_freq)
    else:
        market = "默认"

    df["freq_edt"] = df["dt"].apply(lambda x: freq_end_time(x, target_freq, market))
    dfk1 = df.groupby("freq_edt").agg(
        {
            "symbol": "first",
            "dt": "last",
            "open": "first",
            "close": "last",
            "high": "max",
            "low": "min",
            "vol": "sum",
            "amount": "sum",
            "freq_edt": "last",
        }
    )
    dfk1.reset_index(drop=True, inplace=True)
    dfk1["dt"] = dfk1["freq_edt"]
    dfk1 = dfk1[["symbol", "dt", "open", "close", "high", "low", "vol", "amount"]]

    if raw_bars:
        _bars = []
        for i, row in enumerate(dfk1.to_dict("records"), 1):
            row.update({"id": i, "freq": target_freq})
            _bars.append(RawBar(**row))

        if kwargs.get("drop_unfinished", True):
            # 清除最后一根未完成的K线
            if df["dt"].iloc[-1] < _bars[-1].dt:
                _bars.pop()
        return _bars
    else:
        return dfk1


class BarGenerator:

    version = "V231008"

    def __init__(self, base_freq: str, freqs: List[str], max_count: int = 5000, market="默认"):
        self.symbol = None
        self.end_dt = None
        self.market = market
        self.base_freq = base_freq
        self.max_count = max_count
        self.freqs = freqs
        self.bars = {v: [] for v in self.freqs}
        self.bars.update({base_freq: []})
        self.freq_map = {f.value: f for _, f in Freq.__members__.items()}
        self.__validate_freqs()

    def __validate_freqs(self):
        from czsc.utils import sorted_freqs

        if self.base_freq not in sorted_freqs:
            raise ValueError(f"base_freq is not in sorted_freqs: {self.base_freq}")

        i = sorted_freqs.index(self.base_freq)
        f = sorted_freqs[i:]
        for freq in self.freqs:
            if freq not in f:
                raise ValueError(f"freqs中包含不支持的周期：{freq}")

    def init_freq_bars(self, freq: str, bars: List[RawBar]):
        """初始化某个周期的K线序列

        函数计算逻辑：

        1. 首先，它断言`freq`必须是`self.bars`的键之一。如果`freq`不在`self.bars`的键中，代码会抛出一个断言错误。
        2. 然后，它断言`self.bars[freq]`必须为空。如果`self.bars[freq]`不为空，代码会抛出一个断言错误，并显示一条错误消息。
        3. 如果以上两个断言都通过，它会将`bars`赋值给`self.bars[freq]`，从而初始化指定频率的K线序列。
        4. 最后，它会将`bars`列表中的最后一个`RawBar`对象的`symbol`属性赋值给`self.symbol`。

        :param freq: 周期名称
        :param bars: K线序列
        """
        assert freq in self.bars.keys()
        assert not self.bars[freq], f"self.bars['{freq}'] 不为空，不允许执行初始化"
        self.bars[freq] = bars
        self.symbol = bars[-1].symbol

    def __repr__(self):
        return f"<BarGenerator for {self.symbol} @ {self.end_dt}>"

    def _update_freq(self, bar: RawBar, freq: Freq) -> None:
        """更新指定周期K线

        函数计算逻辑：

        1. 计算目标频率的结束时间`freq_edt`。
        2. 检查`self.bars`中是否已经有目标频率的K线。如果没有，创建一个新的`RawBar`对象，并将其添加到`self.bars`中，然后返回。
        3. 如果已经有目标频率的K线，获取最后一根K线`last`。
        4. 检查`freq_edt`是否不等于最后一根K线的日期时间。如果不等于，创建一个新的`RawBar`对象，并将其添加到`self.bars`中。
        5. 如果`freq_edt`等于最后一根K线的日期时间，创建一个新的`RawBar`对象，其开盘价为最后一根K线的开盘价，
            收盘价为当前K线的收盘价，最高价为最后一根K线和当前K线的最高价中的最大值，最低价为最后一根K线和当前K线的最低价中的最小值，
            成交量和成交金额为最后一根K线和当前K线的成交量和成交金额的和。然后用这个新的`RawBar`对象替换`self.bars`中的最后一根K线。

        :param bar: 基础周期已完成K线
        :param freq: 目标周期
        """
        freq_edt = freq_end_time(bar.dt, freq, self.market)

        if not self.bars[freq.value]:
            bar_ = RawBar(
                symbol=bar.symbol,
                freq=freq,
                dt=freq_edt,
                id=0,
                open=bar.open,
                close=bar.close,
                high=bar.high,
                low=bar.low,
                vol=bar.vol,
                amount=bar.amount,
            )
            self.bars[freq.value].append(bar_)
            return

        last: RawBar = self.bars[freq.value][-1]
        if freq_edt != self.bars[freq.value][-1].dt:
            bar_ = RawBar(
                symbol=bar.symbol,
                freq=freq,
                dt=freq_edt,
                id=last.id + 1,
                open=bar.open,
                close=bar.close,
                high=bar.high,
                low=bar.low,
                vol=bar.vol,
                amount=bar.amount,
            )
            self.bars[freq.value].append(bar_)

        else:
            bar_ = RawBar(
                symbol=bar.symbol,
                freq=freq,
                dt=freq_edt,
                id=last.id,
                open=last.open,
                close=bar.close,
                high=max(last.high, bar.high),
                low=min(last.low, bar.low),
                vol=last.vol + bar.vol,
                amount=last.amount + bar.amount,
            )
            self.bars[freq.value][-1] = bar_

    def update(self, bar: RawBar) -> None:
        """更新各周期K线

        函数计算逻辑：

        1. 首先，它获取基准频率`base_freq`，并断言`bar`的频率值等于`base_freq`。
        2. 然后，它将`bar`的符号和日期时间设置为`self.symbol`和`self.end_dt`。
        3. 接下来，它检查是否已经有一个与`bar`日期时间相同的K线存在于`self.bars[base_freq]`中。
            如果存在，它会记录一个警告并返回，不进行任何更新。
        4. 如果不存在重复的K线，它会遍历`self.bars`的所有键（即所有的频率），并对每个频率调用`self._update_freq`方法来更新该频率的K线。
        5. 最后，它会限制在内存中的K线数量，确保每个频率的K线数量不超过`self.max_count`。

        :param bar: 必须是已经结束的Bar
        :return: None
        """
        base_freq = self.base_freq
        if bar.freq.value != base_freq:
            raise ValueError(
                f"Input bar frequency does not match base frequency. Expected {base_freq}, got {bar.freq.value}"
            )
        self.symbol = bar.symbol
        self.end_dt = bar.dt

        if self.bars[base_freq] and self.bars[base_freq][-1].dt == bar.dt:
            logger.warning(
                f"BarGenerator.update: 输入重复K线，基准周期为{base_freq}; \n\n输入K线为{bar};\n\n 上一根K线为{self.bars[base_freq][-1]}"
            )
            return

        for freq in self.bars.keys():
            self._update_freq(bar, self.freq_map[freq])

        # 限制存在内存中的K限制数量
        for f, b in self.bars.items():
            if len(b) > self.max_count:
                self.bars[f] = b[-self.max_count :]



# coding: utf-8
import os
import zipfile
from tqdm import tqdm
import pandas as pd
from czsc.analyze import CZSC, RawBar, NewBar, remove_include, FX, check_fx, Direction, kline_pro
from czsc.enum import Freq
from collections import OrderedDict


cur_path = os.path.split(os.path.realpath(__file__))[0]


def read_1min():
    with zipfile.ZipFile(os.path.join(cur_path, 'data/000001.XSHG_1min.zip'), 'r') as z:
        f = z.open('000001.XSHG_1min.csv')
        data = pd.read_csv(f, encoding='utf-8')

    data['dt'] = pd.to_datetime(data['dt'])
    data['amount'] = data['close'] * data['vol']
    records = data.to_dict('records')

    bars = []
    for row in tqdm(records, desc='read_1min'):
        bar = RawBar(**row)
        bar.freq = Freq.F1
        bars.append(bar)
    return bars


def read_daily():
    file_kline = os.path.join(cur_path, "data/000001.SH_D.csv")
    kline = pd.read_csv(file_kline, encoding="utf-8")
    kline['amount'] = kline['close'] * kline['vol']

    kline.loc[:, "dt"] = pd.to_datetime(kline.dt)
    bars = [RawBar(symbol=row['symbol'], id=i, freq=Freq.D, open=row['open'], dt=row['dt'],
                   close=row['close'], high=row['high'], low=row['low'], vol=row['vol'], amount=row['amount'])
            for i, row in kline.iterrows()]
    return bars


def test_find_bi():
    bars = read_daily()
    # 去除包含关系
    bars1 = []
    for bar in bars:
        if len(bars1) < 2:
            bars1.append(NewBar(symbol=bar.symbol, id=bar.id, freq=bar.freq,
                                dt=bar.dt, open=bar.open,
                                close=bar.close, high=bar.high, low=bar.low,
                                vol=bar.vol, amount=bar.amount, elements=[bar]))
        else:
            k1, k2 = bars1[-2:]
            has_include, k3 = remove_include(k1, k2, bar)
            if has_include:
                bars1[-1] = k3
            else:
                bars1.append(k3)

    fxs = []
    for i in range(1, len(bars1) - 1):
        fx = check_fx(bars1[i - 1], bars1[i], bars1[i + 1])
        if isinstance(fx, FX):
            fxs.append(fx)


def test_czsc_update():
    bars = read_daily()
    # 不计算任何信号
    c = CZSC(bars)
    assert not c.signals

    # 测试 ubi 属性
    ubi = c.ubi
    assert ubi['direction'] == Direction.Down
    assert ubi['high_bar'].dt < ubi['low_bar'].dt
    # 测试自定义信号
    c = CZSC(bars, get_signals=None)

    # # 查看每一笔的加速度
    # for bi in c.bi_list:
    #     print(bi.fx_a.dt, bi.fx_b.dt)
    #     print(bi.slope)

    kline = [x.__dict__ for x in c.bars_raw]
    bi = [{'dt': x.fx_a.dt, "bi": x.fx_a.fx} for x in c.bi_list] + \
         [{'dt': c.bi_list[-1].fx_b.dt, "bi": c.bi_list[-1].fx_b.fx}]
    chart = kline_pro(kline, bi=bi, title="{} - {}".format(c.symbol, c.freq))
    file_html = "x.html"
    chart.render(file_html)
    os.remove(file_html)



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2024/7/26 15:46
describe: 资产组合研究相关的工具函数
"""
import loguru


def max_sharp(df, weight_bounds=(0, 1), **kwargs):
    """最大夏普比例组合

    依赖 PyPortfolioOpt 库，需要安装 pip install PyPortfolioOpt

    :param df: pd.DataFrame, 包含多个资产的日收益率数据，index 为日期
    :param weight_bounds: tuple, 权重范围, 默认 (0, 1), 参考 EfficientFrontier 的 weight_bounds 参数
        如果需要设置不同的权重范围，可以传入 dict，如 {"asset1": (0, 0.5), "asset2": (0.1, 0.5)}
    :param kwargs: 其他参数

        - logger: loguru.logger, 默认 None, 日志记录器
        - rounding: int, 默认 4, 权重四舍五入的小数位数

    :return: dict, 资产权重
    """
    from pypfopt import EfficientFrontier
    from pypfopt import expected_returns
    from pypfopt import risk_models
    from czsc.utils.stats import daily_performance

    logger = kwargs.get("logger", loguru.logger)
    df = df.copy()
    if "dt" in df.columns:
        df = df.set_index("dt")

    mu = expected_returns.mean_historical_return(df, returns_data=True)
    S = risk_models.risk_matrix(df, returns_data=True, method="sample_cov")

    if isinstance(weight_bounds, dict):
        weight_bounds = [
            [weight_bounds.get(asset, (0, 1))[0] for asset in df.columns],
            [weight_bounds.get(asset, (0, 1))[1] for asset in df.columns],
        ]

    # Optimize for maximal Sharpe ratio
    ef = EfficientFrontier(mu, S, weight_bounds=weight_bounds)
    _ = ef.max_sharpe()
    weights = ef.clean_weights(cutoff=1e-4, rounding=kwargs.get("rounding", 4))

    logger.info(f"资产权重：{weights}")

    portfolio = df.apply(lambda x: x * weights[x.name], axis=0)
    stats = daily_performance(portfolio.sum(axis=1).to_list())
    logger.info(f"组合表现：{stats}")

    return weights
